
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Antisymmetrization of components of a function in
// interpolating wavelet basis.

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "AntisymmetrizeComponents.h"
#include "ArrayUtils.h"
#include "OwnError.h"

ColumnVector
AntisymmetrizeComponents(
  ColumnVector vFunction,
  Array< Integer > const anDecSof,
  bool fZPresentInBasis
  )
  {
  Integer const nParts = anDecSof.length();
  Integer const nBasis = vFunction.length();
  if( nParts == 0 )
    {
    if( nBasis == 0 )
      {
      return ColumnVector();
      }
    else
      {
      OwnError( "AntisymmetrizeComponents: error in basis" );
      return ColumnVector();
      }
    }
  else
    {
    // vFnA filled with zero
    // No need to set Skmin,i=0 to zero
    ColumnVector vFnA( nBasis, 0.0 );
    Integer const nS0 = anDecSof( 0 );
    if( fZPresentInBasis )
      {
      if( nS0 % 2 == 0 )
        {
        OwnError( "AntisymmetrizeComponents: invalid basis" );
        return ColumnVector();
        }
      if( nS0 < 1 || nS0 > nBasis )
        {
        OwnError( "AntisymmetrizeComponents: invalid basis decomposition" );
        return ColumnVector();
        }
      Integer const n = ( nS0 - 1 ) / 2;
      for( Integer i = 0; i < n; i++ )
        {
        Real const rCompNeg = vFunction( i );
        Real const rCompPos = vFunction( 2 * n - i );
        Real const rComp = ( rCompPos - rCompNeg ) / 2;
        vFnA( i ) = -rComp;
        vFnA( 2 * n - i ) = rComp;
        }
      }
    else
      {
      if( nS0 % 2 != 0 )
        {
        OwnError( "AntisymmetrizeComponents: invalid basis" );
        return ColumnVector();
        }
      if( nS0 < 0 || nS0 > nBasis )
        {
        OwnError( "AntisymmetrizeComponents: invalid basis decomposition" );
        return ColumnVector();
        }
      Integer const n = nS0 / 2;
      for( Integer i = 0; i < n; i++ )
        {
        Real const rCompNeg = vFunction( i );
        Real const rCompPos = vFunction( 2 * n - i - 1 );
        Real const rComp = ( rCompPos - rCompNeg ) / 2;
        vFnA( i ) = -rComp;
        vFnA( 2 * n - i - 1 ) = rComp;
        }
      }
    for( Integer iPart = 1; iPart < nParts; iPart++ )
      {
      Integer const nPart = anDecSof( iPart );
      if( nPart % 2 != 0 || nPart < 0 )
        {
        OwnError( "AntisymmetrizeComponents: invalid basis decomposition" );
        return ColumnVector();
        }
      Integer const n = nPart / 2;
      Integer const iBase =
        SumArrayElementsInRange( anDecSof, Integer( 0 ), 0, iPart - 1 );
      if( iBase + 2 * n > nBasis )
        {
        OwnError( "AntisymmetrizeComponents: index out of range" );
        return ColumnVector();
        }
      for( Integer i = 0; i < n; i++ )
        {
        Real const rCompNeg = vFunction( iBase + i );
        Real const rCompPos = vFunction( iBase + 2 * n - 1 - i );
        Real const rComp = ( rCompPos - rCompNeg ) / 2;
        vFnA( iBase + i ) = -rComp;
        vFnA( iBase + 2 * n - 1 - i ) = rComp;
        }
      }
    return vFnA;
    }
  }
