
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Basis class

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "Basis.h"
#include "ArrayUtils.h"

Basis::Basis( Integer count, Integer** aiData )
  : aIndices( count ),
    fValid( true )
  {
  for( Integer p = 0; p < count; p++ )
    {
    aIndices( p ) = BasisMultiIndex( 
      aiData[ p ][ 0 ],
      aiData[ p ][ 1 ],
      aiData[ p ][ 2 ]
      );
    if( ( aIndices( p ).wt != 0 && aIndices( p ).wt != 1 ) ||
        ( aIndices( p ).si < 0 ) )
      {
      fValid = false;
      }
    }
  }
  
Basis& Basis::operator = ( Basis const & basis )
  {
  if( this != &basis )
    {
    fValid = basis.fValid;
    aIndices = basis.aIndices;
    }
  return *this;
  }

ScalingIndex Basis::GetMinResolution() const
  {
  Integer const n = aIndices.length();
  ScalingIndex siMin = aIndices( 0 ).si;
  for( Integer i = 1; i < n; i++ )
    {
    if( aIndices( i ).si < siMin )
      {
      siMin = aIndices( i ).si;
      }
    }
  return siMin;
  }

ScalingIndex Basis::GetMaxResolution() const
  {
  Integer const n = aIndices.length();
  ScalingIndex siMax = aIndices( n - 1 ).si;
  for( Integer i = n - 2; i >= 0; i-- )
    {
    if( aIndices( i ).si > siMax )
      {
      siMax = aIndices( i ).si;
      }
    }
  return siMax;
  }

Array< Integer > Basis::GetSofDecomposition() const
  {
  Array< Integer > aiDec( 0 );
  Integer const nBasis = length();
  if( nBasis > 0 )
    {
    Integer nCount = 1;
    if( (*this)( 0 ).wt == 1 )
      {
      // If SOF basis has no S part,
      // first element in decomposition is zero.
      aiDec = Array< Integer >( 1, 0 );
      }
    for( Integer imi = 1; imi < nBasis; imi++ )
      {
      WaveletType wt1 = (*this)( imi - 1 ).wt;
      ScalingIndex si1 = (*this)( imi - 1 ).si;
      WaveletType wt2 = (*this)( imi ).wt;
      ScalingIndex si2 = (*this)( imi ).si;
      if( wt1 != wt2 || si1 != si2 )
        {
        if( wt2 == 1 )
          {
          if( si1 <= si2 )
            {
            aiDec = ConcatenateArrays(
              aiDec,
              Array< Integer >( 1, nCount )
              );
            nCount = 0;
            if( wt1 != wt2 && si1 != si2 )
              {
              aiDec = ConcatenateArrays(
                aiDec,
                Array< Integer >( si2 - si1, 0 )
                );
              }
            else if( wt1 == wt2 && si1 != si2 )
              {
              if( si2 - si1 > 1 )
                {
                aiDec = ConcatenateArrays(
                  aiDec,
                  Array< Integer >( si2 - si1 - 1, 0 )
                  );
                }
              }
            }
          else
            {
            OwnError( "Basis::GetSofDecomposition: basis has to contain scaling indices in non-descending order" );
            return Array< Integer >();
            }
          }
        else
          {
          OwnError( "Basis::GetSofDecomposition: basis has to contain only one S part (1)" );
          return Array< Integer >();
          }
        }
      nCount++;
      }
    if( nCount != 0 )
      {
      aiDec = ConcatenateArrays(
        aiDec,
        Array< Integer >( 1, nCount )
        );
      }
    }
  if( SumArrayElements( aiDec, Integer( 0 ) ) != nBasis )
    {
    OwnError( "Basis::GetSofDecomposition: internal error" );
    return Array< Integer >();
    }
  return aiDec;
  }

Array< Integer > Basis::GetNsofDecomposition() const
  {
  Array< Integer > aiDec( 0 );
  Integer const nBasis = length();
  if( nBasis > 0 )
    {
    Integer nCount = 1;
    for( Integer imi = 1; imi < nBasis; imi++ )
      {
      WaveletType wt1 = (*this)( imi - 1 ).wt;
      ScalingIndex si1 = (*this)( imi - 1 ).si;
      WaveletType wt2 = (*this)( imi ).wt;
      ScalingIndex si2 = (*this)( imi ).si;
      if( wt1 != wt2 || si1 != si2 )
        {
        if( si2 >= si1 )
          {
          aiDec = ConcatenateArrays(
            aiDec,
            Array< Integer >( 1, nCount )
            );
          nCount = 0;
          if( wt2 == 0 )
            {
            if( si2 > si1 + 1 )
              {
              aiDec = ConcatenateArrays(
                aiDec,
                Array< Integer >( 2 * ( si2 - si1 - 1 ) )
                );
              }
            }
          else
            {
            if( si2 > si1 )
              {
              aiDec = ConcatenateArrays(
                aiDec,
                Array< Integer >( 2 * ( si2 - si1 ) )
                );
              }
            }
          }
        else
          {
          OwnError( "Basis::GetNsofDecomposition: scaling indices have to be in non-descending order" );
          return Array< Integer >();
          }
        }
      else if( wt1 != wt2 && si1 == si2 )
        {
        if( wt2 == 0 )
          {
          OwnError( "Basis::GetNsofDecomposition: Sk follows Dk" );
          return Array< Integer >();
          }
        aiDec = ConcatenateArrays(
          aiDec,
          Array< Integer >( 1, nCount )
          );
        nCount = 0;
        }
      else if( wt1 == wt2 && si1 != si2 )
        {
        if( si2 > si1 )
          {
          aiDec = ConcatenateArrays(
            aiDec,
            Array< Integer >( 1, nCount )
            );
          nCount = 0;
          aiDec = ConcatenateArrays(
            aiDec,
            Array< Integer >( 2 * ( si2 - si1 ) - 1, 0 )
            );
          }
        else
          {
          OwnError( "Basis::GetNsofDecomposition: scaling indices have to be in non-descending order (2)" );
          return Array< Integer >();
          }
        }
      nCount++;
      }
    if( nCount != 0 )
      {
      aiDec = ConcatenateArrays(
        aiDec,
        Array< Integer >( 1, nCount )
        );
      }
    }
  if( SumArrayElements( aiDec, Integer( 0 ) ) != nBasis )
    {
    OwnError( "Basis::GetNsofDecomposition: internal error" );
    return Array< Integer >();
    }
  return aiDec;
  }

Basis Basis::ConstructBasisWithZ( Integer iIndexToZ ) const
  {
  Integer const nLength = aIndices.length();
  if( iIndexToZ < 0 || iIndexToZ >= nLength )
    {
    OwnError( "Basis::ConstructBasisWithZ: index out of range" );
    return Basis();
    }
  TranslationIndex const siMin = GetMinResolution();
  return Basis(
    ConcatenateArrays(
      ConcatenateArrays(
        GetRangeFromArray( aIndices, 0, iIndexToZ - 1 ),
        Array< BasisMultiIndex >( 1, BasisMultiIndex( 0, siMin, 0 ) )
        ),
      GetRangeFromArray( aIndices, iIndexToZ, nLength - 1 )
      )
    );
  }
