
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Binary input/output for real matrices

// Header file

// Copyright (C) 2004  Tommi Hynlnmaa

#include "BinaryIO.h"

static
int16 ExtractInt16( byte an[] )
  {
#ifdef BINARY_IO_BIG_ENDIAN
  int16 const nResult =
    int16( an[ 1 ] ) + ( int16( an[ 0 ] ) << 8 );
  fprintf( stderr, "BE\n" );
#else
  int16 const nResult =
    int16( an[ 0 ] ) + ( int16( an[ 1 ] ) << 8 );
#endif
  return nResult;
  }

static
void StoreInt16( byte an[], int16 n )
  {
#ifdef BINARY_IO_BIG_ENDIAN
  an[ 1 ] = ( n & 0xFF );
  an[ 0 ] = ( ( n >> 8 ) & 0xFF );
#else
  an[ 0 ] = ( n & 0xFF );
  an[ 1 ] = ( ( n >> 8 ) & 0xFF );
#endif
  }

static
int16 SwapInt16( int16 n )
  {
  return ( ( n & 0xFF ) << 8 ) | ( ( n >> 8 ) & 0xFF );
  }

static
int32 ExtractInt32( byte an[] )
  {
#ifdef BINARY_IO_BIG_ENDIAN
  int32 const nResult =
    int32( an[ 3 ] )
    + ( int32( an[ 2 ] ) << 8 )
    + ( int32( an[ 1 ] ) << 16 )
    + ( int32( an[ 0 ] ) << 24 );
#else
  int32 const nResult =
    int32( an[ 0 ] )
    + ( int32( an[ 1 ] ) << 8 )
    + ( int32( an[ 2 ] ) << 16 )
    + ( int32( an[ 3 ] ) << 24 );
#endif
  return nResult;
  }

static
int32 SwapInt32( int32 n )
  {
  return 
    ( ( n & 0xFF ) << 24 ) 
    | ( ( ( n >> 8 ) & 0xFF ) << 16 )
    | ( ( ( n >> 16 ) & 0xFF ) << 8 )
    | ( ( n >> 24 ) & 0xFF );
  }

static
void StoreInt32( byte an[], int16 n )
  {
#ifdef BINARY_IO_BIG_ENDIAN
  an[ 3 ] = ( n & 0xFF );
  an[ 2 ] = ( ( n >> 8 ) & 0xFF );
  an[ 1 ] = ( ( n >> 16 ) & 0xFF );
  an[ 0 ] = ( ( n >> 24 ) & 0xFF );
#else
  an[ 0 ] = ( n & 0xFF );
  an[ 1 ] = ( ( n >> 8 ) & 0xFF );
  an[ 2 ] = ( ( n >> 16 ) & 0xFF );
  an[ 3 ] = ( ( n >> 24 ) & 0xFF );
#endif
  }

void WriteDoubleBin( FILE* file, double value )
  {
#ifndef ENDIAN_SWAP_BINARY_OUTPUT
  fwrite( &value, sizeof( double ), 1, file );
#else
  int i;
  unsigned char* p;
  p = ( unsigned char* ) &value;
  for( i = 0; i < sizeof( double ); i++ )
    fputc( p[ sizeof( double ) - i - 1 ], file );
#endif
  }

void WriteInt32Bin( FILE* file, int32 value )
  {
#ifndef ENDIAN_SWAP_BINARY_OUTPUT
  fwrite( &value, sizeof( int32 ), 1, file );
#else
  int i;
  unsigned char* p;
  p = ( unsigned char* ) &value;
  for( i = 0; i < sizeof( int32 ); i++ )
    fputc( p[ sizeof( int32 ) - i - 1 ], file );
#endif
  }

void WriteInt16Bin( FILE* file, int16 value )
  {
#ifndef ENDIAN_SWAP_BINARY_OUTPUT
  fwrite( &value, sizeof( int16 ), 1, file );
#else
  int i;
  unsigned char* p;
  p = ( unsigned char* ) &value;
  for( i = 0; i < sizeof( int16 ); i++ )
    fputc( p[ sizeof( int16 ) - i - 1 ], file );
#endif
  }

EnumWriteStatus WriteMatrixBin( 
  FILE* file, Matrix mx
  )
  {
  if( file == NULL || ferror( file ) )
    {
    return WRITE_ERROR;
    }

  int rows = mx.rows();
  int columns = mx.columns();

/* Write signature. */
  WriteInt16Bin( file, nSignature );

/* Write store order (by columns). */
/* Store order by columns as liboctave stores matrices. */
  WriteInt16Bin( file, 1 );

/* Write floating point data type size. */
  WriteInt32Bin( file, sizeof( double ) );

/* Write number of rows. */
  WriteInt32Bin( file, rows );
  
/* Write number of columns. */
  WriteInt32Bin( file, columns );

  if( ferror( file ) )
    {
    return WRITE_ERROR;
    }

  fwrite( mx.data(), sizeof( double ), rows * columns, file );

  return ferror( file ) ? WRITE_ERROR : WRITE_OK;
  }

EnumWriteStatus OutputMatrixBin( string sFileName, Matrix mx )
  {
  FILE* file = fopen( sFileName.c_str(), "w" );
  if( file == NULL )
    {
    return WRITE_ERROR;
    }
  EnumWriteStatus nResult = WriteMatrixBin( file, mx );
  fclose( file );
  return nResult;
  }

pair< Matrix, EnumReadStatus > ReadMatrixBin( FILE* file )
  {
  if( file == NULL || ferror( file ) )
    {
    return pair< Matrix, EnumReadStatus >(
      Matrix(),
      READ_ERROR
      );
    }
  byte anHeader[ nHeaderSize ];
  size_t const nReadHeader =
    fread( anHeader, sizeof( byte ), nHeaderSize, file );
  if( int16( nReadHeader ) != nHeaderSize || ferror( file ) )
    {
    return pair< Matrix, EnumReadStatus >(
      Matrix(),
      READ_ERROR
      );
    }
  int16 const nReadSignature =
    ExtractInt16( &anHeader[ 0 ] );
  int16 const nStoreOrder =
    ExtractInt16( &anHeader[ 2 ] );
  int32 const nElementSize =
    ExtractInt32( &anHeader[ 4 ] );
  int32 const nRows =
    ExtractInt32( &anHeader[ 8 ] );
  int32 const nColumns =
    ExtractInt32( &anHeader[ 12 ] );
#ifdef DEBUG_OUTPUT
  for( Integer iB = 0; iB < nHeaderSize; iB ++ )
    {
    fprintf( stderr, "anHeader[ %d ] = %d\n", iB, anHeader[ iB ] );
    }
  fprintf( stderr, "nReadSignature = %d\n", nReadSignature );
  fprintf( stderr, "nStoreOrder = %d\n", nStoreOrder );
  fprintf( stderr, "nElementSize = %d\n", nElementSize );
  fprintf( stderr, "nRows = %d\n", nRows );
  fprintf( stderr, "nColumns = %d\n", nColumns );
#endif
  if( nReadSignature != nSignature )
    {
    // error in signature
    return pair< Matrix, EnumReadStatus >(
      Matrix(),
      READ_ERROR
      );
    }
  if( nRows < 0 || nColumns < 0 )
    {
    // negative dimension(s)
    return pair< Matrix, EnumReadStatus >(
      Matrix(),
      READ_ERROR
      );
    }
  if( nRows == 0 || nColumns == 0 )
    {
    // matrix with zero dimension(s)
    return pair< Matrix, EnumReadStatus >(
      Matrix( nRows, nColumns ),
      READ_OK
      );
    }
  int32 nElements = nRows * nColumns;
  if( nElements <= 0 )
    {
    // integer overflow
    return pair< Matrix, EnumReadStatus >(
      Matrix(),
      READ_ERROR
      );
    }
#ifdef DEBUG_OUTPUT
  fprintf( stderr, "nElements = %d\n", nElements );
#endif
  // endianness of data not checked
  // conversion between architectures not supported
  if( nElementSize == sizeof( double ) )
    {
    double * arData = new double[ nElements ];
    if( arData == NULL )
      {
      // memory allocation error
      return pair< Matrix, EnumReadStatus >(
        Matrix(),
        READ_ERROR
        );
      }
    size_t const nReadData =
      fread( arData, sizeof( double ), nElements, file );
#ifdef DEBUG_OUTPUT
    fprintf( stderr, "nReadData = %d\n", nReadData );
#endif
    if( int32( nReadData ) != nElements || ferror( file ) )
      {
      // read error
      return pair< Matrix, EnumReadStatus >(
        Matrix(),
        READ_ERROR
        );
      }
    Matrix mxResult( nRows, nColumns );
    if( nStoreOrder == 0 )
      {
      // stored by rows
      for( int32 iRow = 0; iRow < nRows; iRow++ )
        {
        for( int32 iColumn = 0; iColumn < nColumns; iColumn++ )
          {
          mxResult( iRow, iColumn ) = arData[ iRow * nColumns + iColumn ];
          }
        }
      }
    else
      {
      // stored by columns
      for( int32 iRow = 0; iRow < nRows; iRow++ )
        {
        for( int32 iColumn = 0; iColumn < nColumns; iColumn++ )
          {
          mxResult( iRow, iColumn ) = arData[ iColumn * nRows + iRow ];
          }
        }
      }
    delete [] arData;
    return pair< Matrix, EnumReadStatus >(
      mxResult,
      READ_OK
      );
    }
  else if( nElementSize == sizeof( long double ) )
    {
    long double * arData = new long double[ nElements ];
    if( arData == NULL )
      {
      // memory allocation error
      return pair< Matrix, EnumReadStatus >(
        Matrix(),
        READ_ERROR
        );
      }
    size_t const nReadData =
      fread( arData, sizeof( long double ), nElements, file );
    if( int32( nReadData ) != nElements || ferror( file ) )
      {
      // read error
      return pair< Matrix, EnumReadStatus >(
        Matrix(),
        READ_ERROR
        );
      }
    Matrix mxResult( nRows, nColumns );
    if( nStoreOrder == 0 )
      {
      // stored by rows
      for( int32 iRow = 0; iRow < nRows; iRow++ )
        {
        for( int32 iColumn = 0; iColumn < nColumns; iColumn++ )
          {
          mxResult( iRow, iColumn ) = 
            double( arData[ iRow * nColumns + iColumn ] );
          }
        }
      }
    else
      {
      // stored by columns
      for( int32 iRow = 0; iRow < nRows; iRow++ )
        {
        for( int32 iColumn = 0; iColumn < nColumns; iColumn++ )
          {
          mxResult( iRow, iColumn ) =
            double( arData[ iColumn * nRows + iRow ] );
          }
        }
      }
    delete [] arData;
    return pair< Matrix, EnumReadStatus >(
      mxResult,
      READ_OK
      );
    }
  else if( nElementSize == sizeof( float ) )
    {
    float * arData = new float[ nElements ];
    if( arData == NULL )
      {
      // memory allocation error
      return pair< Matrix, EnumReadStatus >(
        Matrix(),
        READ_ERROR
        );
      }
    size_t const nReadData =
      fread( arData, sizeof( float ), nElements, file );
    if( int32( nReadData ) != nElements || ferror( file ) )
      {
      // read error
      return pair< Matrix, EnumReadStatus >(
        Matrix(),
        READ_ERROR
        );
      }
    Matrix mxResult( nRows, nColumns );
    if( nStoreOrder == 0 )
      {
      // stored by rows
      for( int32 iRow = 0; iRow < nRows; iRow++ )
        {
        for( int32 iColumn = 0; iColumn < nColumns; iColumn++ )
          {
          mxResult( iRow, iColumn ) = 
            float( arData[ iRow * nColumns + iColumn ] );
          }
        }
      }
    else
      {
      // stored by columns
      for( int32 iRow = 0; iRow < nRows; iRow++ )
        {
        for( int32 iColumn = 0; iColumn < nColumns; iColumn++ )
          {
          mxResult( iRow, iColumn ) =
            float( arData[ iColumn * nRows + iRow ] );
          }
        }
      }
    delete [] arData;
    return pair< Matrix, EnumReadStatus >(
      mxResult,
      READ_OK
      );
    }
  else
    {
    // unsupported floating point format
    return pair< Matrix, EnumReadStatus >(
      Matrix(),
      READ_ERROR
      );
    }
  }

pair< Matrix, EnumReadStatus > InputMatrixBin( string sFileName )
  {
  FILE* file = fopen( sFileName.c_str(), "r" );
  if( file == NULL )
    {
#ifdef DEBUG_OUTPUT
    fprintf( stderr, "Error opening file.\n" );
#endif
    return pair< Matrix, EnumReadStatus >( Matrix(), READ_ERROR );
    }
  pair< Matrix, EnumReadStatus > const mx_rdst =
    ReadMatrixBin( file );
  fclose( file );
  return mx_rdst;
  }

bool EndianSwapBin( FILE* file1, FILE* file2 )
  {
  byte anHeaderR[ nHeaderSize ];
  byte anHeaderW[ nHeaderSize ];
  size_t const nReadHeader =
    fread( anHeaderR, sizeof( byte ), nHeaderSize, file1 );
  if( nReadHeader != size_t( nHeaderSize ) )
    {
    return false;
    }
  int16 const nSignature = 
    ExtractInt16( &anHeaderR[ 0 ] );
  int16 const nStoreOrder = 
    ExtractInt16( &anHeaderR[ 2 ] );
  int32 const nElementSize =
    ExtractInt32( &anHeaderR[ 4 ] );
  int32 const nRows =
    ExtractInt32( &anHeaderR[ 8 ] );
  int32 const nColumns =
    ExtractInt32( &anHeaderR[ 12 ] );
  StoreInt16( &anHeaderW[ 0 ], SwapInt16( nSignature ) );
  StoreInt16( &anHeaderW[ 2 ], SwapInt16( nStoreOrder ) );
  StoreInt32( &anHeaderW[ 4 ], SwapInt32( nElementSize ) );
  StoreInt32( &anHeaderW[ 8 ], SwapInt32( nRows ) );
  StoreInt32( &anHeaderW[ 12 ], SwapInt32( nColumns ) );
  size_t const nWroteHeader =
    fwrite( anHeaderW, sizeof( byte ), nHeaderSize, file2 );
  if( nWroteHeader != size_t( nHeaderSize ) )
    {
    return false;
    }
  int32 nNumberOfElements = nRows * nColumns;
  for( int32 n = 0; n < nNumberOfElements; n++ )
    {
    byte anR[ sizeof( double ) ];
    byte anW[ sizeof( double ) ];
    size_t const nR =
      fread( anR, sizeof( byte ), sizeof( double ), file1 );
    if( nR != sizeof( double ) )
      {
      return false;
      }
    for( int32 nD = 0; nD < int32( sizeof( double ) ); nD++ )
      {
      anW[ sizeof( double ) - 1 - nD ] = anR[ nD ];
      }
    size_t const nW =
      fwrite( anW, sizeof( byte ), sizeof( double ), file2 );
    if( nW != sizeof( double ) )
      {
      return false;
      }
    }
  return true;
  }
