
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Choosing wavefunction with desired principal
// and angular momentum quantum numbers.

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "ChooseWavefunction.h"
#include "ConvertVectorToNearestScfBasis.h"
#include "FilterValues.h"
#include "ComputeNormalizationCoefficient.h"
#include "AntisymmetrizeComponents.h"
#include "CountWavefunctionNodes.h"
#include "ChooseSignForWavefunction.h"
#include "ArrayUtils.h"
#include "NTwoToIntegerPower.h"

// #define DEBUG_OUTPUT

#ifdef DEBUG_OUTPUT
#include <iostream>
#endif

WavefunctionInfo ChooseWavefunction(
  WaveletFilters const & wv,
  Array< Integer > anDecSof,
  bool fZPresentInBasis,
  Basis const & basisNsof,
  Array< Integer > anDecNsof,
  UniformBasis const & usfbScfBasis,
  ColumnVector vD1Filter,
  ComplexColumnVector cvEigenvalues,
  ComplexMatrix cmxEigenvectors,
  Integer nDesiredNodes,
  Real rAntisymmThreshold,
  Real rDerivThreshold,
  Real rSignCheckPosition
  )
  {
  bool fFound = false;
  Array< Integer > anCountedNodes( 0 );
  Array< Integer > anCountedRejectedNodes( 0 );
  ColumnVector vAntisymmIndicators;
  Array< EnumWfRejectionReason > anRejectionReasons( 0 );
  Integer const nBasis = cvEigenvalues.length();
  if( cmxEigenvectors.rows() != nBasis ||
      cmxEigenvectors.columns() != nBasis )
    {
    OwnError( "ChooseWavefunction: error in eigenproblem solution" );
    return WavefunctionInfo();
    }
  WavefunctionInfo wfinfo;
  wfinfo.rEpsilon = 0.0;
  Integer const nIntegrationDiv =
    NTwoToIntegerPower( usfbScfBasis.si );
  Integer iWavefunction = 0;
  for( iWavefunction = 0;
       iWavefunction < nBasis;
       iWavefunction++ )
    {
    ComplexColumnVector const cvWfCurr =
      cmxEigenvectors.column( iWavefunction );
    Complex const cxEpsilonCurr = cvEigenvalues( iWavefunction );
    anCountedNodes = ConcatenateArrays(
      anCountedNodes,
      Array< Integer >( 1, -1 )
      );
    anCountedRejectedNodes = ConcatenateArrays(
      anCountedRejectedNodes,
      Array< Integer >( 1, -1 )
      );
    vAntisymmIndicators =
      vAntisymmIndicators.stack( ColumnVector( 1, 0.0 ) );
    anRejectionReasons =
      ConcatenateArrays(
        anRejectionReasons,
        Array< EnumWfRejectionReason >( 1, rejNotRejected )
        );
    // imaginary parts have to be *exactly* zero
    // maybe should be compared against some threshold
    ColumnVector const vImWfCurr = imag( cvWfCurr );
    Real const rEpsilonCurr = real( cxEpsilonCurr );
    if( imag( cxEpsilonCurr ) != 0.0 || 
        vImWfCurr.transpose() * vImWfCurr != 0.0 )
      {
      anRejectionReasons( iWavefunction ) = rejComplex;
      }
    else if( rEpsilonCurr >= 0.0 )
      {
      anRejectionReasons( iWavefunction ) = rejPositiveEnergy;
      }
    else
      {
      ColumnVector const vWfCurr = real( cvWfCurr );
      ColumnVector const vWfScfNotNormCurr =
        ConvertVectorToNearestScfBasis(
          wv, basisNsof, anDecNsof, usfbScfBasis, vWfCurr
          );
      ColumnVector const vWfScfCurr =
        vWfScfNotNormCurr /
        ComputeNormalizationCoefficient(
          usfbScfBasis, vWfScfNotNormCurr
          );
      ColumnVector const vWfA =
        AntisymmetrizeComponents(
          vWfCurr, anDecSof, fZPresentInBasis
          );
      ColumnVector const vWfAScf =
        ConvertVectorToNearestScfBasis(
          wv, basisNsof, anDecNsof, usfbScfBasis, vWfA
          );
#ifdef DEBUG_OUTPUT
      cout << "N(vWfCurr) = " << vWfCurr.length() << endl;
      cout << "N(vWfA) = " << vWfA.length() << endl;
      cout << "WF SOF component and antisymmetrized component:" << endl;
      for( Integer i = 0; i < vWfCurr.length(); i++ )
        {
        cout << vWfCurr( i ) << ", " << vWfA( i ) << endl;
        }
      cout << "N(vWfScfCurr) = " << vWfScfCurr.length() << endl;
      cout << "N(vWfAScf) = " << vWfAScf.length() << endl;
      cout << "WF usfb component and antisymmetrized component:" << endl;
      for( Integer i = 0; i < vWfScfCurr.length(); i++ )
        {
        cout << vWfScfCurr( i ) << ", " << vWfAScf( i ) << endl;
        }
#endif
      Real const rNormCoeff =
        ComputeNormalizationCoefficient( usfbScfBasis, vWfAScf );
      ColumnVector const vWfAN = vWfA / rNormCoeff;
      ColumnVector const vWfAScfN = vWfAScf / rNormCoeff;
      ColumnVector const vAntisymmDiff = vWfAScfN - vWfScfCurr;
      Real const rAntisymmIndicator =
        ( vAntisymmDiff.transpose() * vAntisymmDiff )
        / nIntegrationDiv;
      vAntisymmIndicators( iWavefunction ) = rAntisymmIndicator;
      if( rAntisymmIndicator > rAntisymmThreshold )
        {
#ifdef DEBUG_OUTPUT
        cout << "ChooseWavefunction: rejecting nonantisymmetric solution: "
             << endl
             << "  index = " << iWavefunction << endl
             << "  antisymmetricity indicator = " << rAntisymmIndicator 
             << endl
             << "  antisymmetricity threshold = " << rAntisymmThreshold
             << endl;
#endif
        anRejectionReasons( iWavefunction ) = rejNotAntisymmetric;
        }
      else
        {
        ColumnVector vWfDerivScf =
          Real( nIntegrationDiv )
          * FilterValues( vWfAScfN, vD1Filter );
        pair< Integer, Integer > const nNodes_nRejNodes =
          CountWavefunctionNodes(
            usfbScfBasis, vWfAScfN, vWfDerivScf, rDerivThreshold
            );
        Integer const nNodes = nNodes_nRejNodes.first;
        anCountedNodes( iWavefunction ) = nNodes;
        anCountedRejectedNodes( iWavefunction ) = nNodes_nRejNodes.second;
        if( nNodes == nDesiredNodes )
          {
          fFound = true;
          Integer const nWfSign = ChooseSignForWavefunction(
            usfbScfBasis, vWfAScfN, rSignCheckPosition
            );
          wfinfo.vWavefunction =
            Real( nWfSign ) * vWfAN;
          wfinfo.vWfScf =
            Real( nWfSign ) * vWfAScfN;
          wfinfo.rEpsilon = rEpsilonCurr;
          wfinfo.iIndexFound = iWavefunction;
          // terminate for-loop
          break;
          }
        else
          {
          anRejectionReasons( iWavefunction ) = rejWrongNumberOfNodes;
          }
        }
      }
    }
  wfinfo.fFound = fFound;
  wfinfo.nCounted = iWavefunction;
  wfinfo.anRejectionReasons = anRejectionReasons;
  wfinfo.anCountedNodes = anCountedNodes;
  wfinfo.anCountedRejectedNodes = anCountedRejectedNodes;
  wfinfo.vAntisymmIndicators = vAntisymmIndicators;
  return wfinfo;
  }

