
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of atom energy in Hartree-Fock method.

// Header file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "ComputeAtomEnergy.h"
#include "ArrayUtils.h"
#include "NTwoToIntegerPower.h"
#include "GetNonnegativePart.h"
#include "ConvertVectorToNearestScfBasis.h"

Triplet< Real, ColumnVector, ColumnVector >
ComputeAtomEnergy(
  WaveletFilters const & wv,
  Basis const & basisSof,
  Basis const & basisNsof,
  Array< Integer > anDecNsof,
  UniformBasis const & usfb,
  Array< Integer > anOrbitalOccupations,
  Matrix mxWavefunctions,
  ColumnVector vDiagLagrangeMultipliers,
  Array< Matrix > amxDirectIntOps,
  Array< Matrix > amxExchangeIntOps
  )
  {
  Integer const nOrbitals = anOrbitalOccupations.length();
  if( vDiagLagrangeMultipliers.length() != nOrbitals )
    {
    OwnError( "ComputeAtomEnergy: error in argument vDiagLagrangeMultipliers" );
    return Triplet< Real, ColumnVector, ColumnVector >(
      0.0, ColumnVector(), ColumnVector()
      );
    }
  if( amxDirectIntOps.length() != nOrbitals )
    {
    OwnError( "ComputeAtomEnergy: error in argument amxDirectIntOps" );
    return Triplet< Real, ColumnVector, ColumnVector >(
      0.0, ColumnVector(), ColumnVector()
      );
    }
  if( amxExchangeIntOps.length() != nOrbitals )
    {
    OwnError( "ComputeAtomEnergy: error in argument amxExchangeIntOps" );
    return Triplet< Real, ColumnVector, ColumnVector >(
      0.0, ColumnVector(), ColumnVector()
      );
    }
  ColumnVector vJTerms( nOrbitals, 0.0 );
  ColumnVector vKTerms( nOrbitals, 0.0 );
  Integer const nIntDiv = NTwoToIntegerPower( usfb.si );
  for( Integer iOrbital = 0; iOrbital < nOrbitals; iOrbital++ )
    {
    Matrix const mxJCurr = amxDirectIntOps( iOrbital );
    Matrix const mxKCurr = amxExchangeIntOps( iOrbital );
    ColumnVector const vWfCurr = mxWavefunctions.column( iOrbital );
    ColumnVector const vWfCurrScf =
      ConvertVectorToNearestScfBasis(
        wv, basisNsof, anDecNsof, usfb, vWfCurr
        );
    pair< UniformBasis, ColumnVector > const usfb_vWfNonneg =
      GetNonnegativePart(
        pair< UniformBasis, ColumnVector >( usfb, vWfCurrScf )
        );
    ColumnVector const vWfNonneg = usfb_vWfNonneg.second;
    ColumnVector const vJWf = mxJCurr * vWfCurr;
    ColumnVector const vJWfScf =
      ConvertVectorToNearestScfBasis(
        wv, basisNsof, anDecNsof, usfb, vJWf
        );
    pair< UniformBasis, ColumnVector > const usfb_vJWfNonneg =
      GetNonnegativePart(
        pair< UniformBasis, ColumnVector >( usfb, vJWfScf )
        );
    ColumnVector const vJWfNonneg = usfb_vJWfNonneg.second;
    ColumnVector const vKWf = mxKCurr * vWfCurr;
    ColumnVector const vKWfScf =
      ConvertVectorToNearestScfBasis(
        wv, basisNsof, anDecNsof, usfb, vKWf
        );
    pair< UniformBasis, ColumnVector > const usfb_vKWfNonneg =
      GetNonnegativePart(
        pair< UniformBasis, ColumnVector >( usfb, vKWfScf )
        );
    ColumnVector const vKWfNonneg = usfb_vKWfNonneg.second;
    vJTerms( iOrbital ) =
      ( vWfNonneg.transpose() * vJWfNonneg ) / nIntDiv;
    vKTerms( iOrbital ) =
      ( vWfNonneg.transpose() * vKWfNonneg ) / nIntDiv;
    }
  ColumnVector vOrbitalOccupations( nOrbitals, 0.0 );
  for( Integer iOrbital = 0; iOrbital < nOrbitals; iOrbital++ )
    {
    vOrbitalOccupations( iOrbital ) =
      Real( anOrbitalOccupations( iOrbital ) );
    }
  Real const rEnergy =
    SumArrayElements(
      ( Array< Real > )
      product( 
        vOrbitalOccupations,
        vDiagLagrangeMultipliers - 0.5 * ( vJTerms - vKTerms )
        ),
      Real( 0.0 )
      );
  return Triplet< Real, ColumnVector, ColumnVector >(
    rEnergy, vJTerms, vKTerms
    );
  }
