
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of NSOF block for differentiation operator
//
//   (Op f)(r) = (df^k/dr^k)(r)
//

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "ComputeNsofBlockDerivOp.h"
#include "NTwoToIntegerPower.h"
#include "OwnError.h"

static
Matrix ComputeNsofBlockDerivOpSS(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector filter,
  Integer order
  )
  {
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      Integer index = j - i;
      if( index >= -wv.nM && index <= wv.nM )
        {
        a( p, q ) = filter( index + wv.nM );
        }
      }
    }
  return Real( NTwoToIntegerPower( k * order ) ) * a;
  }

static
Matrix ComputeNsofBlockDerivOpSD(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector filter,
  Integer order
  )
  {
  Integer const nM = wv.nM;
  ColumnVector const vHt = wv.vHt;
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      Integer index = 2 * j - 2 * i + 1;
      if( index >= -nM && index <= nM )
	{
        a( p, q ) = - filter( index + nM );
	}
      }
    }
  return Real( NTwoToIntegerPower( ( k + 1 ) * order) ) * a;
  }

static
Matrix ComputeNsofBlockDerivOpDS(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector filter,
  Integer order
  )
  {
  Integer const nM = wv.nM;
  ColumnVector const vH = wv.vH;
  ColumnVector const vGt = wv.vGt;
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      for( Integer nu = -nM; nu <= nM; nu++ )
        {
        for( Integer mu = -nM; mu <= nM; mu++ )
          {
          Integer index = 2 * j - 2 * i + mu - nu;
          if( index >= -nM && index <= nM )
            {
            a( p, q ) += 
              vGt( nu + nM ) * 
              vH( mu + nM ) *
              filter( index + nM );
            }
          }
        }
      }
    }
  return Real( NTwoToIntegerPower( ( k + 1 ) * order ) ) * a;
  }

static
Matrix ComputeNsofBlockDerivOpDD( 
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector filter,
  Integer order
  )
  {
  Integer const nM = wv.nM;
  ColumnVector const vGt = wv.vGt;
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      for( Integer nu = -nM; nu <= nM; nu++ )
        {
        Integer index = 2 * j - 2 * i + 1 - nu;
        if( index >= -nM && index <= nM )
          {
          a( p, q ) -= vGt( nu + nM ) * filter( index + nM );
          }
        }
      }
    }
  return Real( NTwoToIntegerPower( ( k + 1 ) * order ) ) * a;
  }

Matrix ComputeNsofBlockDerivOp( 
  WaveletFilters const & wv,
  ScalingIndex k,
  WaveletType wtTarget,
  WaveletType wtSource,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector filter,
  Integer order
  )
  {
  if( k + 2 > MAX_RESOLUTION ||
      order < 0 ||
      filter.length() != 2 * wv.nM + 1 )
    {
    OwnError( "ComputeNsofBlockDerivOp: error in arguments." );
    return Matrix( 0, 0 );
    }
  Matrix result;
  if( wtTarget == 0 && wtSource == 0 )
    {
    result = ComputeNsofBlockDerivOpSS( 
      wv, k, aiTarget, aiSource,
      filter, order
      );
    }
  else if( wtTarget == 0 && wtSource == 1 )
    {
    result = ComputeNsofBlockDerivOpSD( 
      wv, k, aiTarget, aiSource,
      filter, order
      );
    }
  else if( wtTarget == 1 && wtSource == 0 )
    {
    result = ComputeNsofBlockDerivOpDS( 
      wv, k, aiTarget, aiSource,
      filter, order
      );
    }
  else if( wtTarget == 1 && wtSource == 1 )
    {
    result = ComputeNsofBlockDerivOpDD( 
      wv, k, aiTarget, aiSource,
      filter, order
      );
    }
  else
    {
    OwnError( "ComputeNsofBlockDerivOp: internal error" );
    result = Matrix( 0, 0 );
    }
  return result;
  }
