
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of NSOF block for integral-to-infinity operator.

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "ComputeNsofBlockIntegralInf.h"
#include "NTwoToIntegerPower.h"
#include "OwnError.h"
#include <cmath>

static
Matrix ComputeNsofBlockIntegralInfSS(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< Integer > aiTarget,
  Array< Integer > aiSource,
  MotherIntegral const & mother_integral
  )
  {
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = TranslationIndex( aiTarget( p ) );
      TranslationIndex j = TranslationIndex( aiSource( q ) );
      a( p, q ) =
        ( mother_integral.GetValueAtInf()
          - mother_integral.GetValueAt( 2 * ( abs( i ) - j ) ) )
        / NTwoToIntegerPower( k );
      }
    }
  return a;
  }

static
Matrix ComputeNsofBlockIntegralInfSD(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< Integer > aiTarget,
  Array< Integer > aiSource,
  MotherIntegral const & mother_integral
  )
  {
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = TranslationIndex( aiTarget( p ) );
      TranslationIndex j = TranslationIndex( aiSource( q ) );
      Integer n1 = 2 * abs( i ) - 2 * j - 1;
      a( p, q ) =
        -( mother_integral.GetValueAtInf()
           - mother_integral.GetValueAt( 2 * n1 ) )
        / NTwoToIntegerPower( k + 1 );
      }
    }
  return a;
  }

static
Matrix ComputeNsofBlockIntegralInfDS(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< Integer > aiTarget,
  Array< Integer > aiSource,
  MotherIntegral const & mother_integral
  )
  {
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = TranslationIndex( aiTarget( p ) );
      TranslationIndex j = TranslationIndex( aiSource( q ) );
      for( Integer nu = -wv.nM; nu <= wv.nM; nu++ )
        {
        Integer index = abs( 2 * i + nu ) - 2 * j;
        Real r1 = mother_integral.GetValueAt( index );
        a( p, q ) += wv.vGt( wv.nM + nu ) 
          * ( mother_integral.GetValueAtInf() - r1 );
        }
      }
    }
  return a / Real( NTwoToIntegerPower( k ) );
  }

static
Matrix ComputeNsofBlockIntegralInfDD(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< Integer > aiTarget,
  Array< Integer > aiSource,
  MotherIntegral const & mother_integral
  )
  {
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = TranslationIndex( aiTarget( p ) );
      TranslationIndex j = TranslationIndex( aiSource( q ) );
      for( Integer nu = -wv.nM; nu <= wv.nM; nu++ )
        {
        Integer index = 2 * ( abs( 2 * i + nu ) - 2 * j - 1 );
        Real r1 = mother_integral.GetValueAt( index );
        a( p, q ) += wv.vGt( wv.nM + nu ) 
          * ( mother_integral.GetValueAtInf() - r1 );
        }
      }
    }
  return -a / Real( NTwoToIntegerPower( k + 1 ) );
  }

Matrix ComputeNsofBlockIntegralInf(
  WaveletFilters const & wv,
  ScalingIndex k,
  WaveletType wtTarget,
  WaveletType wtSource,     
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  MotherIntegral const & mother_integral
  )
  {
  if( k < 0 )
    {
    OwnError( "ComputeNsofBlockIntegralInf: error in argument k" );
    return Matrix( 0, 0 );
    }
  if( wtTarget != 0 && wtTarget != 1 )
    {
    OwnError( "ComputeNsofBlockIntegralInf: error in argument wtTarget" );
    return Matrix( 0, 0 );
    }
  if( wtSource != 0 && wtSource != 1 )
    {
    OwnError( "ComputeNsofBlockIntegralInf: error in argument wtSource" );
    return Matrix( 0, 0 );
    }
  if( k + 1 > MAX_RESOLUTION )
    {
    OwnError( "ComputeNsofBlockIntegralInf: error in argument k (overflow)" );
    return Matrix( 0, 0 );
    }
  if( mother_integral.GetFunction().usfb.si != 1 )
    {
    OwnError( "ComputeNsofBlockIntegralZero: error in mother integral" );
    return Matrix( 0, 0 );
    }
  Matrix result;
  if( wtTarget == 0 && wtSource == 0 )
    {
    result = ComputeNsofBlockIntegralInfSS(
      wv, k, aiTarget, aiSource,
      mother_integral
      );
    }
  else if( wtTarget == 0 && wtSource == 1 )
    {
    result = ComputeNsofBlockIntegralInfSD(
      wv, k, aiTarget, aiSource,
      mother_integral
      );
    }
  else if( wtTarget == 1 && wtSource == 0 )
    {
    result = ComputeNsofBlockIntegralInfDS(
      wv, k, aiTarget, aiSource,
      mother_integral
      );
    }
  else if( wtTarget == 1 && wtSource == 1 )
    {
    result = ComputeNsofBlockIntegralInfDD(
      wv, k, aiTarget, aiSource,
      mother_integral
      );
    }
  else
    {
    OwnError( "ComputeNsofBlockIntegralInf: internal error" );
    result = Matrix( 0, 0 );
    }
  return result;
  }
