
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of NSOF block for operator
//
//   (R_{-1} f)(r) 
//
//       / f(r) / |r|;   |r| >= c
//     = <
//       \ f(r) / c;   |r| < c
//

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "ComputeNsofBlockInvR.h"
#include "NTwoToIntegerPower.h"
#include "OwnError.h"

static inline
Real InverseR( Real r, Real c )
  {
  if( fabs( r ) >= c )
    {
    return 1.0 / fabs( r );
    }
  else
    {
    return 1.0 / c;
    }
  }

static
Matrix ComputeNsofBlockInvRSS(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  Real rCutPoint 
  )
  {
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      if( i == j )
        {
	Real x = Real( i ) / NTwoToIntegerPower( k );
        a( p, q ) = InverseR( x, rCutPoint );
        }
      }
    }
  return a;
  }

static
Matrix ComputeNsofBlockInvRSD(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  Real rCutPoint 
  )
  {
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  return a;
  }

static
Matrix ComputeNsofBlockInvRDS(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  Real rCutPoint 
  )
  {
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      Integer index = 2 * i - 2 * j + 1;
      if( index >= -wv.nM && index <= wv.nM )
        {
        Real filter_value = wv.vH( wv.nM + index );
        Real x1 = Real( j ) / NTwoToIntegerPower( k );
        Real x2 = Real( 2 * i + 1 )
          / NTwoToIntegerPower( k + 1 );
        a( p, q ) = 
          filter_value 
          * ( InverseR( x1, rCutPoint ) - InverseR( x2, rCutPoint ) );
        }
      }
    }
  return a;
  }

static
Matrix ComputeNsofBlockInvRDD(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  Real rCutPoint 
  )
  {
  Integer rows = aiTarget.length();
  Integer columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      if( i == j )
        {
        Real x = Real( 2 * i + 1 )
          / NTwoToIntegerPower( k + 1 );
        a( p, q ) = InverseR( x, rCutPoint );
        }
      }
    }
  return a;
  }

Matrix ComputeNsofBlockInvR( 
  WaveletFilters const & wv,
  ScalingIndex k,
  WaveletType wtTarget,
  WaveletType wtSource,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  Real rCutPoint
  )
  {
  if( k + 2 > MAX_RESOLUTION ||
      rCutPoint < 0.0 )
    {
    OwnError( "ComputeNsofBlockInvR: error in arguments." );
    return Matrix( 0, 0 );
    }
  Matrix result;
  if( wtTarget == 0 && wtSource == 0 )
    {
    result = ComputeNsofBlockInvRSS( 
      wv, k, aiTarget, aiSource,
      rCutPoint
      );
    }
  else if( wtTarget == 0 && wtSource == 1 )
    {
    result = ComputeNsofBlockInvRSD( 
      wv, k, aiTarget, aiSource,
      rCutPoint
      );
    }
  else if( wtTarget == 1 && wtSource == 0 )
    {
    result = ComputeNsofBlockInvRDS( 
      wv, k, aiTarget, aiSource,
      rCutPoint
      );
    }
  else if( wtTarget == 1 && wtSource == 1 )
    {
    result = ComputeNsofBlockInvRDD( 
      wv, k, aiTarget, aiSource,
      rCutPoint
      );
    }
  else
    {
    OwnError( "ComputeNsofBlockInvR: internal error" );
    result = Matrix( 0, 0 );
    }
  return result;
  }
