
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of NSOF block for Laplacian operator
//
//   (Op f)(r) = (df^2/dr^2)(r)
//

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "ComputeNsofBlockLaplacian.h"
#include "NTwoToIntegerPower.h"
#include "OwnError.h"

// #define DEBUG_OUTPUT

#ifdef DEBUG_OUTPUT
#include "BinaryIO.h"
#endif

static
Matrix ComputeNsofBlockLaplacianSS(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector filter
  )
  {
  Integer const n = ( filter.length() - 1 ) / 2;
  Integer const rows = aiTarget.length();
  Integer const columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      Integer index = j - i;
      if( index >= -n && index <= n )
        {
        a( p, q ) = filter( index + n );
        }
      }
    }
  return Real( NTwoToIntegerPower( 2 * k ) ) * a;
  }

static
Matrix ComputeNsofBlockLaplacianSD(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector filter
  )
  {
  Integer const n = ( filter.length() - 1 ) / 2;
  Integer const rows = aiTarget.length();
  Integer const columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      Integer index = 2 * j - 2 * i + 1;
      if( index >= -n && index <= n )
	{
        a( p, q ) = - filter( index + n );
	}
      }
    }
  return Real( 4 * NTwoToIntegerPower( 2 * k ) ) * a;
  }

static
Matrix ComputeNsofBlockLaplacianDS(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector filter
  )
  {
  Integer const n = ( filter.length() - 1 ) / 2;
  Integer const nM = wv.nM;
  ColumnVector const vH = wv.vH;
  ColumnVector const vGt = wv.vGt;
  Integer const rows = aiTarget.length();
  Integer const columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      for( Integer nu = -nM; nu <= nM; nu++ )
        {
        for( Integer mu = -nM; mu <= nM; mu++ )
          {
          Integer index = 2 * j - 2 * i + mu - nu;
          if( index >= -n && index <= n )
            {
            a( p, q ) += 
              vGt( nu + nM ) * 
              vH( mu + nM ) *
              filter( index + n );
            }
          }
        }
      }
    }
  return Real( 4 * NTwoToIntegerPower( 2 * k ) ) * a;
  }

static
Matrix ComputeNsofBlockLaplacianDD( 
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector filter
  )
  {
  Integer const n = ( filter.length() - 1 ) / 2;
  Integer const nM = wv.nM;
  ColumnVector const vGt = wv.vGt;
  Integer const rows = aiTarget.length();
  Integer const columns = aiSource.length();
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      for( Integer nu = -nM; nu <= nM; nu++ )
        {
        Integer index = 2 * j - 2 * i + 1 - nu;
        if( index >= -n && index <= n )
          {
          a( p, q ) -= vGt( nu + nM ) * filter( index + n );
          }
        }
      }
    }
  return Real( 4 * NTwoToIntegerPower( 2 * k ) ) * a;
  }

Matrix ComputeNsofBlockLaplacian( 
  WaveletFilters const & wv,
  ScalingIndex k,
  WaveletType wtTarget,
  WaveletType wtSource,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector filter
  )
  {
  if( ( k + 2 > siMaxResolution ) ||
      ( filter.length() < 1 ) ||
      ( filter.length() % 2 == 0 ) )
    {
    OwnError( "ComputeNsofBlockLaplacian: error in arguments." );
    return Matrix( 0, 0 );
    }
  Matrix result;
  if( wtTarget == 0 && wtSource == 0 )
    {
    result = ComputeNsofBlockLaplacianSS( 
      wv, k, aiTarget, aiSource,
      filter
      );
    }
  else if( wtTarget == 0 && wtSource == 1 )
    {
    result = ComputeNsofBlockLaplacianSD( 
      wv, k, aiTarget, aiSource,
      filter
      );
    }
  else if( wtTarget == 1 && wtSource == 0 )
    {
    result = ComputeNsofBlockLaplacianDS( 
      wv, k, aiTarget, aiSource,
      filter
      );
    }
  else if( wtTarget == 1 && wtSource == 1 )
    {
    result = ComputeNsofBlockLaplacianDD( 
      wv, k, aiTarget, aiSource,
      filter
      );
    }
  else
    {
    OwnError( "ComputeNsofBlockLaplacian: internal error" );
    result = Matrix( 0, 0 );
    }
  return result;
  }
