
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of NSOF block for multiplication operator
//
//   (M(f) g)(r) = f(r) g(r)
//

// Header file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "ComputeNsofBlockProduct.h"
#include "NTwoToIntegerPower.h"
#include "OwnError.h"

static
Matrix ComputeNsofBlockProductSS(
  WaveletFilters const & wv,
  Integer k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector vComponents,
  UniformBasis const & usfb
  )
  {
  Integer const z = 
    NTwoToIntegerPower( usfb.si - k - 1 );
  Integer const n = aiTarget.length();
  Integer const m = aiSource.length();
  Matrix result( n, m );
  result.fill( 0.0 );
  for( Integer p = 0; p < n; p++ )
    {
    for( Integer q = 0; q < m; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      if( i == j )
        {
        Integer s = 2 * z * i;
        if( s >= usfb.tiMin && s <= usfb.tiMax )
          {
          result( p, q ) = vComponents( s - usfb.tiMin );
          }
        }
      }
    }
  return result;
  }

static
Matrix ComputeNsofBlockProductSD(
  WaveletFilters const & wv,
  Integer k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector vComponents,
  UniformBasis const & usfb
  )
  {
  Integer const n = aiTarget.length();
  Integer const m = aiSource.length();
  Matrix result( n, m );
  result.fill( 0.0 );
  return result;
  }

static
Matrix ComputeNsofBlockProductDS(
  WaveletFilters const & wv,
  Integer k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector vComponents,
  UniformBasis const & usfb
  )
  {
  Integer const z = 
    NTwoToIntegerPower( usfb.si - k - 1 );
  Integer const n = aiTarget.length();
  Integer const m = aiSource.length();
  Matrix result( n, m );
  result.fill( 0.0 );
  for( Integer p = 0; p < n; p++ )
    {
    for( Integer q = 0; q < m; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      Integer mu = 2 * i - 2 * j + 1;
      if( mu >= -wv.nM && mu <= wv.nM )
        {
        Integer s1 = 2 * z * j;
        Integer s2 = z * ( 2 * i + 1 );
        Real felem = wv.vH( mu + wv.nM );
        Real c1 =
          ( s1 >= usfb.tiMin && s1 <= usfb.tiMax )
          ? vComponents( s1 - usfb.tiMin )
          : 0.0;
        Real c2 =
          ( s2 >= usfb.tiMin && s2 <= usfb.tiMax )
          ? vComponents( s2 - usfb.tiMin )
          : 0.0;
        result( p, q ) = felem * ( c1 - c2 );
        }
      }
    }
  return result;
  }

static
Matrix ComputeNsofBlockProductDD(
  WaveletFilters const & wv,
  Integer k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector vComponents,
  UniformBasis const & usfb
  )
  {
  Integer const z = 
    NTwoToIntegerPower( usfb.si - k - 1 );
  Integer const n = aiTarget.length();
  Integer const m = aiSource.length();
  Matrix result( n, m );
  result.fill( 0.0 );
  for( Integer p = 0; p < n; p++ )
    {
    for( Integer q = 0; q < m; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      if( i == j )
        {
        Integer s = z * ( 2 * i + 1 );
        if( s >= usfb.tiMin && s <= usfb.tiMax )
          {
          result( p, q ) = vComponents( s - usfb.tiMin );
          }
        }
      }
    }
  return result;
  }

Matrix ComputeNsofBlockProduct( 
  WaveletFilters const & wv,
  ScalingIndex k,
  WaveletType wtTarget,
  WaveletType wtSource,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource,
  ColumnVector vComponents,
  UniformBasis const & usfb
  )
  {
  if( k + 2 > siMaxResolution )
    {
    OwnError( "ComputeNsofBlockProduct: error in arguments." );
    return Matrix( 0, 0 );
    }
  if( usfb.wt != 0 )
    {
    OwnError( "ComputeNsofBlockProduct: usfb has to contain scaling functions" );
    return Matrix( 0, 0 );
    }
  if( usfb.tiMin > usfb.tiMax )
    {
    OwnError( "ComputeNsofBlockProduct: error in argument usfb" );
    return Matrix( 0, 0 );
    }
  if( k >= usfb.si )
    {
    OwnError( "ComputeNsofBlockProduct: invalid k value in usfb" );
    return Matrix( 0, 0 );
    }
  Matrix result;
  if( wtTarget == 0 && wtSource == 0 )
    {
    result = ComputeNsofBlockProductSS( 
      wv, k, aiTarget, aiSource,
      vComponents, usfb
      );
    }
  else if( wtTarget == 0 && wtSource == 1 )
    {
    result = ComputeNsofBlockProductSD( 
      wv, k, aiTarget, aiSource,
      vComponents, usfb
      );
    }
  else if( wtTarget == 1 && wtSource == 0 )
    {
    result = ComputeNsofBlockProductDS( 
      wv, k, aiTarget, aiSource,
      vComponents, usfb
      );
    }
  else if( wtTarget == 1 && wtSource == 1 )
    {
    result = ComputeNsofBlockProductDD( 
      wv, k, aiTarget, aiSource,
      vComponents, usfb
      );
    }
  else
    {
    OwnError( "ComputeNsofBlockProduct: internal error" );
    result = Matrix( 0, 0 );
    }
  return result;
  }
