
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of NSOF block for operator
//
//   (Rf)(r) = |r| f(r)
//

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "ComputeNsofBlockR.h"
#include "NTwoToIntegerPower.h"
#include "OwnError.h"

// #define DEBUG_OUTPUT

#ifdef DEBUG_OUTPUT
#include <iostream>
#endif

static
Matrix ComputeNsofBlockRSS(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource
  )
  {
  Integer const rows = aiTarget.length();
  Integer const columns = aiSource.length();
#ifdef DEBUG_OUTPUT
  for( Integer i1 = 0; i1 < rows; i1++ )
    {
    cout << aiTarget( i1 ) << " ";
    }
  cout << endl;
  for( Integer i2 = 0; i2 < columns; i2++ )
    {
    cout << aiSource( i2 ) << " ";
    }
  cout << endl;
#endif
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      if( i == j )
        {
	Real x = Real( i ) / NTwoToIntegerPower( k );
        a( p, q ) = fabs( x );
        }
      }
    }
  return a;
  }

static
Matrix ComputeNsofBlockRSD(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource
  )
  {
  Integer const rows = aiTarget.length();
  Integer const columns = aiSource.length();
#ifdef DEBUG_OUTPUT
  for( Integer i1 = 0; i1 < rows; i1++ )
    {
    cout << aiTarget( i1 ) << " ";
    }
  cout << endl;
  for( Integer i2 = 0; i2 < columns; i2++ )
    {
    cout << aiSource( i2 ) << " ";
    }
  cout << endl;
#endif
  Matrix a( rows, columns );
  a.fill( 0.0 );
  return a;
  }

static
Matrix ComputeNsofBlockRDS(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource
  )
  {
  Integer const rows = aiTarget.length();
  Integer const columns = aiSource.length();
#ifdef DEBUG_OUTPUT
  for( Integer i1 = 0; i1 < rows; i1++ )
    {
    cout << aiTarget( i1 ) << " ";
    }
  cout << endl;
  for( Integer i2 = 0; i2 < columns; i2++ )
    {
    cout << aiSource( i2 ) << " ";
    }
  cout << endl;
#endif
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      Integer index = 2 * i - 2 * j + 1;
      if( index >= -wv.nM && index <= wv.nM )
        {
        Real filter_value = wv.vH( wv.nM + index );
        Real x1 = Real( j ) / NTwoToIntegerPower( k );
        Real x2 = Real( 2 * i + 1 )
          / NTwoToIntegerPower( k + 1 );
        a( p, q ) = 
          filter_value 
          * ( fabs( x1 ) - fabs( x2 ) );
        }
      }
    }
  return a;
  }

static
Matrix ComputeNsofBlockRDD(
  WaveletFilters const & wv,
  ScalingIndex k,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource
  )
  {
  Integer const rows = aiTarget.length();
  Integer const columns = aiSource.length();
#ifdef DEBUG_OUTPUT
  for( Integer i1 = 0; i1 < rows; i1++ )
    {
    cout << aiTarget( i1 ) << " ";
    }
  cout << endl;
  for( Integer i2 = 0; i2 < columns; i2++ )
    {
    cout << aiSource( i2 ) << " ";
    }
  cout << endl;
#endif
  Matrix a( rows, columns );
  a.fill( 0.0 );
  for( Integer p = 0; p < rows; p++ )
    {
    for( Integer q = 0; q < columns; q++ )
      {
      TranslationIndex i = aiTarget( p );
      TranslationIndex j = aiSource( q );
      if( i == j )
        {
        Real x = Real( 2 * i + 1 ) 
          / NTwoToIntegerPower( k + 1 );
        a( p, q ) = fabs( x );
        }
      }
    }
  return a;
  }

Matrix ComputeNsofBlockR( 
  WaveletFilters const & wv,
  ScalingIndex k,
  WaveletType wtTarget,
  WaveletType wtSource,
  Array< TranslationIndex > aiTarget,
  Array< TranslationIndex > aiSource
  )
  {
  if( k + 2 > siMaxResolution )
    {
    OwnError( "ComputeNsofBlockR: error in arguments." );
    return Matrix( 0, 0 );
    }
  Matrix result;
  if( wtTarget == 0 && wtSource == 0 )
    {
    result = ComputeNsofBlockRSS( 
      wv, k, aiTarget, aiSource
      );
    }
  else if( wtTarget == 0 && wtSource == 1 )
    {
    result = ComputeNsofBlockRSD( 
      wv, k, aiTarget, aiSource
      );
    }
  else if( wtTarget == 1 && wtSource == 0 )
    {
    result = ComputeNsofBlockRDS( 
      wv, k, aiTarget, aiSource
      );
    }
  else if( wtTarget == 1 && wtSource == 1 )
    {
    result = ComputeNsofBlockRDD( 
      wv, k, aiTarget, aiSource
      );
    }
  else
    {
    OwnError( "ComputeNsofBlockR: internal error" );
    result = Matrix( 0, 0 );
    }
  return result;
  }
