
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of matrix of an operator in SOF given a function that
// calculates NSOF blocks of the operator.

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "ComputeOperator.h"
#include "BasisPart.h"
#include "GetBasisPartNsof.h"
#include "ComputeOperatorSofPart.h"
#include "RemoveTopFromNsofBasis.h"

#ifdef DEBUG_OUTPUT
#include <iostream>
#endif

Matrix ComputeOperator(
  WaveletFilters const & wv,
  NsofBlockComputator const & nsofblockcomputator,
  Basis const & basisnsof,
  Array< Integer > aiDecNsof
  )
  {
#ifdef DEBUG_OUTPUT
  cout << "ComputeOperator ENTER" << endl;
#endif
  Integer const nLevels = aiDecNsof.length();
  Integer const nBasis = basisnsof.length();
#ifdef DEBUG_OUTPUT
  cout << "  nLevels = " << nLevels << endl;
  cout << "  nBasis = " << nBasis << endl;
#endif
  if( nBasis == 0 )
    {
    OwnError( "ComputeOperator: error in basis" );
    return Matrix( 0, 0 );
    }
  if( nLevels != 1 && nLevels % 2 != 0 )
    {
    OwnError( "ComputeOperator: number of parts in NSOF basis has to be even or one" );
    return Matrix( 0, 0 );
    }
  if( nLevels < 1 )
    {
    OwnError( "ComputeOperator: NSOF basis has to containt at least one part" );
    return Matrix( 0, 0 );
    }
  else if( nLevels == 1 )
    {
    // Only S part in basis
    ScalingIndex const siCurrent = basisnsof( 0 ).si;
    BasisPart const amiS = GetBasisPartNsof(
      basisnsof, aiDecNsof, siCurrent, 0, siCurrent
      );
    Matrix const mxSS = nsofblockcomputator.Evaluate(
      siCurrent, 0, 0, 
      amiS.GetTranslationIndices(),
      amiS.GetTranslationIndices()
      );
    return mxSS;
    }
  else if( nLevels == 2 )
    {
    ScalingIndex const siCurrent = basisnsof( 0 ).si;
    BasisPart const amiS = GetBasisPartNsof(
      basisnsof, aiDecNsof, siCurrent, 0, siCurrent
      );
    BasisPart const amiD = GetBasisPartNsof(
      basisnsof, aiDecNsof, siCurrent, 1, siCurrent
      );
    Matrix const mxSS = nsofblockcomputator.Evaluate(
      siCurrent, 0, 0, 
      amiS.GetTranslationIndices(),
      amiS.GetTranslationIndices()
      );
    Matrix const mxSD = nsofblockcomputator.Evaluate(
      siCurrent, 0, 1, 
      amiS.GetTranslationIndices(),
      amiD.GetTranslationIndices()
      );
    Matrix const mxDS = nsofblockcomputator.Evaluate(
      siCurrent, 1, 0, 
      amiD.GetTranslationIndices(),
      amiS.GetTranslationIndices()
      );
    Matrix const mxDD = nsofblockcomputator.Evaluate(
      siCurrent, 1, 1, 
      amiD.GetTranslationIndices(),
      amiD.GetTranslationIndices()
      );
#ifdef DEBUG_OUTPUT
    cout << "ComputeOperator EXIT 1" << endl;
#endif
    return
      mxSS.append( mxSD ).stack( mxDS.append( mxDD ) );
    }
  else
    {
    ScalingIndex const siMin = basisnsof( 0 ).si;
    ScalingIndex const siCurrent =
      siMin + nLevels / 2 - 1;
#ifdef DEBUG_OUTPUT
    cout << "  siMin = " << siMin << endl;
    cout << "  siCurrent = " << siCurrent << endl;
#endif
    pair< Basis, Array< Integer > > const basis_decPrev =
      RemoveTopFromNsofBasis(
        pair< Basis, Array< Integer > >( basisnsof, aiDecNsof )
        );
    Basis const basisnsofPrev =
      basis_decPrev.first;
    Array< Integer > const aiDecNsofPrev =
      basis_decPrev.second;
    Matrix const mxPrev =
      ComputeOperator(
        wv,
        nsofblockcomputator,
        basisnsofPrev,
        aiDecNsofPrev
        );
    BasisPart const amiS =
      GetBasisPartNsof( basisnsof, aiDecNsof, siMin, 0, siCurrent );
    BasisPart const amiD =
      GetBasisPartNsof( basisnsof, aiDecNsof, siMin, 1, siCurrent );
    Matrix const mxSD =
      nsofblockcomputator.Evaluate(
        siCurrent, 0, 1,
        amiS.GetTranslationIndices(),
        amiD.GetTranslationIndices()
        );
    Matrix const mxDS =
      nsofblockcomputator.Evaluate(
        siCurrent, 1, 0,
        amiD.GetTranslationIndices(),
        amiS.GetTranslationIndices()
        );
    Matrix const mxDD =
      nsofblockcomputator.Evaluate(
        siCurrent, 1, 1,
        amiD.GetTranslationIndices(),
        amiD.GetTranslationIndices()
        );
    BasisPart const amiSKMin =
      GetBasisPartNsof(
        basisnsof, aiDecNsof, siMin, 0, siMin
        );
    Matrix mxVert =
      ComputeOperatorSofPart(
        wv, mxSD, basisnsof, aiDecNsof, 
        siMin,
        0, siMin, 
        amiSKMin.GetTranslationIndices(),
        1, siCurrent,
        amiD.GetTranslationIndices()
        );
    for( Integer q = siMin; q <= siCurrent - 1; q++ )
      {
      BasisPart const amiDCurrent =
        GetBasisPartNsof(
          basisnsof,
          aiDecNsof,
          siMin, 1, q
          );
      Matrix const mxVertCurrent =
        ComputeOperatorSofPart(
          wv, mxSD, basisnsof, aiDecNsof,
          siMin,
          1, q, 
          amiDCurrent.GetTranslationIndices(),
          1, siCurrent,
          amiD.GetTranslationIndices()
          );
      mxVert = mxVert.stack( mxVertCurrent );
      }
    Matrix mxHoriz =
      ComputeOperatorSofPart(
        wv, mxDS, basisnsof, aiDecNsof, 
        siMin,
        1, siCurrent, 
        amiD.GetTranslationIndices(),
        0, siMin,
        amiSKMin.GetTranslationIndices()
        );
    for( Integer q = siMin; q <= siCurrent - 1; q++ )
      {
      BasisPart const amiDCurrent =
        GetBasisPartNsof(
          basisnsof,
          aiDecNsof,
          siMin, 1, q
          );
      Matrix const mxHorizCurrent =
        ComputeOperatorSofPart(
          wv, mxDS, basisnsof, aiDecNsof,
          siMin,
          1, siCurrent, 
          amiD.GetTranslationIndices(),
          1, q,
          amiDCurrent.GetTranslationIndices()
          );
      mxHoriz = mxHoriz.append( mxHorizCurrent );
      }
#ifdef DEBUG_OUTPUT
    cout << "ComputeOperator EXIT 2" << endl;
    cout << "  rows(Prev) = " << mxPrev.rows()
         << ", columns(Prev) = " << mxPrev.columns()
         << endl;
    cout << "  rows(Vert) = " << mxVert.rows()
         << ", columns(Vert) = " << mxVert.columns()
         << endl;
    cout << "  rows(Horiz) = " << mxHoriz.rows()
         << ", columns(Horiz) = " << mxHoriz.columns()
         << endl;
    cout << "  rows(DD) = " << mxDD.rows()
         << ", columns(DD) = " << mxDD.columns()
         << endl;
#endif
    return mxPrev.append( mxVert).stack(
      mxHoriz.append( mxDD )
      );
    }
  }
