
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of Slater integral y values.

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "ComputeSlaterIntegral.h"
#include "RealToNonnegIntPower.h"

#define ASSUME_WF_ANTISYMMETRY

static
ColumnVector ComputeWavefunctionValues(
  ScalingFunction const & scf,
  Basis const & basis,
  ColumnVector wf_comp, 
  Real r_range,
  Integer points
  )
  {
  Real step = r_range / ( points - 1 );
  Integer nbasis = basis.length();
  ColumnVector values( points );
  values.fill( 0.0 );
  for( Integer i = 0; i < points; i++ )
    {
    Real r = step / 2 + i * step;
    for( Integer p = 0; p < nbasis; p++ )
      {
      values( i ) += 
        wf_comp( p ) *
        scf.BasisFunctionValue( basis( p ), r );
      }
    }
  return values;
  }

static
ColumnVector ComputeIntegralFromZero( 
  ColumnVector wf_values1, 
  ColumnVector wf_values2, 
  Integer points,
  Real step,
  Integer slater_k
  )
  {
  ColumnVector values( points );
  values( 0 ) = 0.0;
  for( Integer i = 1; i < points; i++ )
    {
    Real rp = i * step - step / 2;
    values( i ) = 
      values( i - 1 ) +
      RealToNonnegIntPower( rp, slater_k ) *
      wf_values1( i - 1 ) * wf_values2( i - 1 );
    }
  for( Integer i = 0; i < points; i++ )
    {
    values( i ) *= step;
    }
  return values;
  }

// static
// ColumnVector ComputeIntegralFromInf( 
//   ColumnVector wf_values1, 
//   ColumnVector wf_values2, 
//   Integer points,
//   Real step,
//   Integer slater_k
//   )
//   {
//   ColumnVector values( points );
//   values( points - 1 ) = 0.0;
//   for( Integer i = points - 1; i >= 0; i-- )
//     {
//     Real rp = i * step + step / 2;
//     if( i < points - 1 )
//       {
//       values( i ) = values( i + 1 );
//       }
//     else
//       {
//       values( i ) = 0.0;
//       }
//     values( i ) += 
//       ( 1.0 / RealToNonnegIntPower( rp, slater_k + 1 ) ) *
//       wf_values1( i ) * wf_values2( i );
//     }
//   for( Integer i = 0; i < points; i++ ) 
//     {
//     values( i ) *= step;
//     }
//   return values;
//   }

static
ColumnVector ComputeIntegralFromInf( 
  ColumnVector wf_values1, 
  ColumnVector wf_values2, 
  Integer points,
  Real step,
  Integer slater_k
  )
  {
  ColumnVector values( points );
  values( points - 1 ) = 0.0;
  for( Integer i = points - 2; i >= 0; i-- )
    {
    Real rp = i * step + step / 2;
    values( i ) =
      values( i + 1 )
      + ( 1.0 / RealToNonnegIntPower( rp, slater_k + 1 ) )
      * wf_values1( i ) * wf_values2( i );
    }
  for( Integer i = 0; i < points; i++ ) 
    {
    values( i ) *= step;
    }
  return values;
  }

#ifdef ASSUME_WF_ANTISYMMETRY

static
ColumnVector DivideByRPower( 
  ColumnVector values, Real step, int power
  )
  {
  Integer points = values.length();
  ColumnVector result( points );
  // Correct if P_n'l' is antisymmetric
  result( 0 ) = 0.0;
  for( Integer i = 1; i < points; i++ )
    {
    Real r = i * step;
    result( i ) =
      values( i ) / RealToNonnegIntPower( r, power );
    }
  return result;
  }

#else /* ASSUME_WF_ANTISYMMETRY */

static
ColumnVector DivideByRPower( 
  ColumnVector values, Real step, int power
  )
  {
  Integer points = values.length();
  ColumnVector result( points );
  for( Integer i = 1; i < points; i++ )
    {
    Real r = i * step;
    result( i ) = values( i ) / RealToNonnegIntPower( r, power );
    }
  if( power != 0 )
    {
    /* Linear extrapolation for r = 0. */
    result( 0 ) = 2 * result( 1 ) - result( 2 );
    }
  else
    {
    /* If power == 0 point r == 0 is automatically correct. */
    result( 0 ) = values( 0 );
    }
  return result;
  }

#endif /* ASSUME_WF_ANTISYMMETRY */

static
ColumnVector MultiplyByRPower( 
  ColumnVector values, Real step, int power
  )
  {
  Integer points = values.length();
  ColumnVector result( points );
  for( Integer i = 0; i < points; i++ )
    {
    Real r = i * step;
    result( i ) = values( i ) * RealToNonnegIntPower( r, power );
    }
  return result;
  }

ColumnVector DoComputeSlaterIntegral(
  ColumnVector vWfValues1,
  ColumnVector vWfValues2,
  Real rRange,
  Integer nSlaterK,
  bool fNegativePartSignChange,
  Real& rYa,
  Real& rYb
  )
  {
  if( vWfValues1.length() != vWfValues2.length() )
    {
    OwnError( "ComputeSlaterIntegral: internal error" );
    return ColumnVector();
    }
  Integer const nPointsHalf = vWfValues1.length();
  Integer const nPoints = 2 * nPointsHalf - 1;
  Real const rStep = rRange / ( nPointsHalf - 1 );
  ColumnVector const vIntegralFromZero =
    ComputeIntegralFromZero( 
      vWfValues1, vWfValues2,
      nPointsHalf, rStep, nSlaterK
      );
  ColumnVector const vIntegralFromInf =
    ComputeIntegralFromInf( 
      vWfValues1, vWfValues2,
      nPointsHalf, rStep, nSlaterK );
  rYa = vIntegralFromInf( 0 );
  rYb = vIntegralFromZero( vIntegralFromZero.length() - 1 );
  // Y^k( r ), r >= 0
  ColumnVector vSlaterIntPos =
    DivideByRPower( vIntegralFromZero, rStep, nSlaterK + 1 ) +
    MultiplyByRPower( vIntegralFromInf, rStep, nSlaterK );
  ColumnVector vSlaterInt( nPoints );
  // whole Y^k
  for( Integer i = ( nPoints - 1 ) / 2; i < nPoints; i++ )
    {
    vSlaterInt( i ) = vSlaterIntPos( i - ( nPoints - 1 ) / 2 );
    }
  if( fNegativePartSignChange )
    {
    for( Integer i = 0; i < ( nPoints - 1 ) / 2; i++ )
      {
      vSlaterInt( i ) = - vSlaterIntPos( ( nPoints - 1 ) / 2 - i );
      }
    }
  else
    {
    for( Integer i = 0; i < ( nPoints - 1 ) / 2; i++ )
      {
      vSlaterInt( i ) = vSlaterIntPos( ( nPoints - 1 ) / 2 - i );
      }
    }
  return vSlaterInt;
  }

ColumnVector ComputeSlaterIntegral(
  ScalingFunction const & scf,
  Basis const & basis,
  Integer nSlaterK,
  ColumnVector vComponents1,
  ColumnVector vComponents2,
  Real rRange,
  Integer nSlaterIntPoints,
  bool fNegativePartSignChange,
  Real& rYa,
  Real& rYb
  )
  {
  return DoComputeSlaterIntegral(
    ComputeWavefunctionValues(
      scf, basis, vComponents1,
      rRange,
      ( nSlaterIntPoints - 1 ) / 2 + 1
      ),
    ComputeWavefunctionValues(
      scf, basis, vComponents2,
      rRange,
      ( nSlaterIntPoints - 1 ) / 2 + 1
      ),
    rRange,
    nSlaterK,
    fNegativePartSignChange,
    rYa,
    rYb
    );
  }
