
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Counting nodes from a wavefunction represented
// in uniform scaling function basis.

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "CountWavefunctionNodes.h"
#include "GetNonnegativePart.h"
#include "OwnError.h"

pair< Integer, Integer >
CountWavefunctionNodes(
  UniformBasis const & usfb,
  ColumnVector vWfScf,
  ColumnVector vWfDerivScf,
  Real rDerivThreshold
  )
  {
  pair< UniformBasis, ColumnVector > const usfbWfNonneg_vWfNonneg =
    GetNonnegativePart(
      pair< UniformBasis, ColumnVector >(
        usfb,
        vWfScf
        )
      );
  pair< UniformBasis, ColumnVector > const usfbDerivNonneg_vDerivNonneg =
    GetNonnegativePart(
      pair< UniformBasis, ColumnVector >(
        usfb,
        vWfDerivScf
        )
      );
  UniformBasis const & usfbWfNonneg = usfbWfNonneg_vWfNonneg.first;
  ColumnVector const & vWfNonneg = usfbWfNonneg_vWfNonneg.second;
  UniformBasis const & usfbDerivNonneg
    = usfbDerivNonneg_vDerivNonneg.first;
  ColumnVector const & vDerivNonneg
    = usfbDerivNonneg_vDerivNonneg.second;
  if( vWfNonneg.length() != vDerivNonneg.length() )
    {
    OwnError( "CountWavefunctionNodes: internal error" );
    return pair< Integer, Integer >( -1, -1 );
    }
  if( usfbWfNonneg != usfbDerivNonneg )
    {
    OwnError( "CountWavefunctionNodes: internal error" );
    return pair< Integer, Integer >( -1, -1 );
    }
  if( usfbWfNonneg.tiMin > usfbWfNonneg.tiMax )
    {
    return pair< Integer, Integer >( 0, 0 );
    }
  Integer const n = usfbWfNonneg.tiMax - usfbWfNonneg.tiMin + 1;
  if( vWfNonneg.length() != n )
    {
    OwnError( "CountWavefunctionNodes: internal error" );
    return pair< Integer, Integer >( -1, -1 );
    }
  Integer nNodes = 0;
  Integer nRejectedNodes = 0;
  for( Integer i = 1; i <= n - 2; i++ )
    {
    Real const rY1 = vWfNonneg( i );
    Real const rY2 = vWfNonneg( i + 1 );
    if( ( rY1 >= 0 && rY2 < 0 ) ||
        ( rY1 <= 0 && rY2 > 0 ) )
      {
      Real const rD1 = vDerivNonneg( i );
      Real const rD2 = vDerivNonneg( i + 1 );
      // y changes sign => no division by zero
      Real const rX0 = i - rY1 / ( rY2 - rY1 );
      Real const rD = ( ( i + 1 ) - rX0 ) * rD1 + ( rX0 - i ) * rD2;
      if( fabs( rD ) >= rDerivThreshold )
        {
        nNodes++;
        }
      else
        {
        nRejectedNodes++;
        }
      }
    }
  return pair< Integer, Integer >( nNodes, nRejectedNodes );
  }
