
// Distributed under GNU Lesser General Public License
// See file LICENSE

// Copyright (C) 2004  Tommi Hynlnmaa

// Helium HF computation

// Implementation file

#include "DoHeliumHFComputation.h"
#include "ComputeOperator.h"
#include "LaplacianNsofBlockComputator.h"
#include "RNsofBlockComputator.h"
#include "ProductNsofBlockComputator.h"
#include "IntegralZeroNsofBlockComputator.h"
#include "IntegralInfNsofBlockComputator.h"
#include "Triplet.h"
#include "SortWavefunctions.h"
#include "ChooseWavefunction.h"
#include "MatrixUtils.h"
#include "VectorUtils.h"
#include "ConvertVectorToNearestScfBasis.h"
#include "ConvertVectorFromScfToStandardBasis.h"
#include "ComputeNormalizationCoefficient.h"
#include "ComputeNormalInnerProduct.h"
#include "PrintHFIterationInfo.h"
#include <cstdlib>
// liboctave headers
#include "EIG.h"

#ifdef DEBUG_OUTPUT
#include <iostream>
#endif

#ifdef USE_NAMESPACES
using namespace std;
#endif

static Integer const nDesiredNodes1s = 0;

HeliumHFComputationResults DoHeliumHFComputation(
  WaveletFilters const & wv,
  Basis const & basisSof,
  Array< Integer > anDecSof,
  Basis const & basisNsof,
  Array< Integer > anDecNsof,
  Basis const & basisZNsof,
  Array< Integer > anDecZNsof,
  Integer iIndexToZ,
  UniformBasis const & usfbScfBasis,
  MotherIntegral const & motherintegral,
  ColumnVector vD1Filter,
  ColumnVector vD2Filter,
  Integer nAtomicNumber,
  Integer nMaxIterations,
  Real rConvThreshold,
  EnumMatrixNormType nMatrixNormType,
  HFOrbitalComputationParams const & hforbparams1s,
  HFLogger& logger
  )
  {
#ifdef DEBUG_OUTPUT
  cout << "DoHeliumHFComputation ENTER" << endl;
#endif
  Integer const nBasis = basisSof.length();
#ifdef DEBUG_OUTPUT
  cout << "Computing Laplacian..." << endl;
#endif
  Matrix const mxOpLaplacian =
    ComputeOperator(
      wv,
      LaplacianNsofBlockComputator( wv, vD2Filter ),
      basisNsof,
      anDecNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing R operator..." << endl;
#endif
  Matrix const mxOpR =
    ComputeOperator(
      wv,
      RNsofBlockComputator( wv ),
      basisNsof,
      anDecNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing R^(-1)..." << endl;
#endif
  Integer nInversionInfo = 0;
  Matrix const mxOpInvR = mxOpR.inverse( nInversionInfo );
  if( nInversionInfo == -1 )
    {
    HeliumHFComputationResults oResults;
    oResults.errhf = errhfSingularROperator;
    return oResults;
    }
#ifdef DEBUG_OUTPUT
  cout << "Computing R operator in Z basis..." << endl;
#endif
  Matrix const mxOpZR =
    ComputeOperator(
      wv,
      RNsofBlockComputator( wv ),
      basisZNsof,
      anDecZNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing integral-from-zero operator in Z basis..." << endl;
#endif
  Matrix const mxOpZIntZero =
    ComputeOperator(
      wv,
      IntegralZeroNsofBlockComputator( wv, motherintegral ),
      basisZNsof,
      anDecZNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing integral-to-infinity operator in Z basis..." 
       << endl;
#endif
  Matrix const mxOpZIntInf =
    ComputeOperator(
      wv,
      IntegralInfNsofBlockComputator( wv, motherintegral ),
      basisZNsof,
      anDecZNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing single-electron Hamiltonian..." << endl;
#endif
  Matrix const mxOpT = (-0.5) * mxOpLaplacian;
  Matrix const mxOpV = ( -nAtomicNumber ) * mxOpInvR;
  Matrix const mxOpH0L0 = mxOpT + mxOpV;
#ifdef DEBUG_OUTPUT
  cout << "Computing Slater integral operator..." << endl;
#endif
  Matrix const mxOp10R = 
    RemoveRowFromMatrix( mxOpZR, iIndexToZ );
  Matrix const mxOp01IntInf = 
    RemoveColumnFromMatrix( mxOpZIntInf, iIndexToZ );
  Matrix const mxOp01IntZero =
    RemoveColumnFromMatrix( mxOpZIntZero, iIndexToZ );
  Matrix const mxOp11IntZero =
    RemoveRowFromMatrix( mxOp01IntZero, iIndexToZ );
#ifdef DEBUG_OUTPUT
  cout << mxOpInvR.rows() << ", " << mxOpInvR.columns() << endl;
  cout << mxOp11IntZero.rows() << ", " << mxOp11IntZero.columns()
       << endl;
#endif
  Matrix const mxOp11Slater0FromZero = mxOpInvR * mxOp11IntZero;
  Matrix const mxOp00Slater0FromZero =
    InsertZToMatrix( mxOp11Slater0FromZero, iIndexToZ );
#ifdef DEBUG_OUTPUT
  cout << mxOp01IntInf.rows() << ", " << mxOp01IntInf.columns()
       << endl;
#endif
  Matrix const mxOp01Slater0ToInf = mxOp01IntInf * mxOpInvR;
  Matrix const mxOp00Slater0ToInf =
    InsertColumnToMatrix(
      mxOp01Slater0ToInf,
      ColumnVector( mxOp01Slater0ToInf.rows(), 0.0 ),
      iIndexToZ
      );
  Matrix const mxOpZSlater0 = mxOp00Slater0FromZero + mxOp00Slater0ToInf;
#ifdef DEBUG_OUTPUT
  cout << " ||S0||_1 = " << ComputeMatrixNorm( mxOpZSlater0, mxnormOne ) << endl;
#endif
#ifdef DEBUG_OUTPUT
  cout << "Preparing HF iteration..." << endl;
#endif
  Matrix mxOpF1sPrev( nBasis, nBasis, 0.0 );
  Matrix mxOpF1s( nBasis, nBasis, 0.0 );
  Matrix mxOpJ1s( nBasis, nBasis, 0.0 );
  ColumnVector vWf1s( nBasis, 0.0 );
  ColumnVector vWf1sHydr( nBasis, 0.0 );
  Real rEpsilon1s = 0.0;
  Real rEpsilon1sHydr = 0.0;
  bool fConvergenceObtained = false;
#ifdef DEBUG_OUTPUT
  cout << "Hartree-Fock iteration..." << endl;
#endif
  for( Integer nIteration = 0; 
       nIteration < nMaxIterations; 
       nIteration++ )
    {
    logger.NotifyIterationStep( nIteration );
#ifdef DEBUG_OUTPUT
    cout << "Computing Fock operator" << endl;
#endif
    mxOpF1s = mxOpH0L0 + mxOpJ1s;
#ifdef DEBUG_OUTPUT
    cout << "Solving eigenproblem for 1s..." << endl;
#endif
    EIG eigenproblem1s( mxOpF1s );
    ComplexColumnVector const cvEigenvalues1s =
      eigenproblem1s.eigenvalues();
    ComplexMatrix const cmxEigenvectors1s =
      eigenproblem1s.eigenvectors();
    Triplet< ComplexColumnVector, ComplexMatrix, Array< Integer > > const
      cv_cmx_ai1s =
      SortWavefunctions( cvEigenvalues1s, cmxEigenvectors1s );
    ComplexColumnVector const & cvSortedEigenvalues1s =
      cv_cmx_ai1s.first;
    ComplexMatrix const & cmxSortedEigenvectors1s =
      cv_cmx_ai1s.second;
    WavefunctionInfo const wfinfo1s =
      ChooseWavefunction(
        wv,
        anDecSof,
        false,
        basisNsof,
        anDecNsof,
        usfbScfBasis,
        vD1Filter,
        cvSortedEigenvalues1s,
        cmxSortedEigenvectors1s,
        nDesiredNodes1s,
        hforbparams1s.rAntisymmThreshold,
        hforbparams1s.rDerivThreshold,
        hforbparams1s.rSignCheckPosition
        );
    if( ! wfinfo1s.fFound )
      {
#ifdef DEBUG_OUTPUT
      cout << "1s Wavefunction not found" << endl;
#endif
      HeliumHFComputationResults oResults;
      oResults.errhf = errhfWavefunctionNotFound;
      oResults.nErrorInOrbital = 0;
      oResults.mxOpLaplacian = mxOpLaplacian;
      oResults.mxOpR = mxOpR;
      oResults.mxOpInvR = mxOpInvR;
      oResults.mxOpZR = mxOpZR;
      oResults.mxOpZIntZero = mxOpZIntZero;
      oResults.mxOpZIntInf = mxOpZIntInf;
      oResults.mxOpZSlater0 = mxOpZSlater0;
      oResults.mxOpH0L0 = mxOpH0L0;
      oResults.mxOpJ1s = mxOpJ1s;
      return oResults;
      }
    rEpsilon1s = wfinfo1s.rEpsilon;
    logger.NotifyOrbitalInfo(
      0, 1, 0, rEpsilon1s, wfinfo1s.anRejectionReasons
      );
#ifdef DEBUG_OUTPUT
    cout << "Computing final wavefunction for current step..." 
         << endl;
#endif
#ifdef DEBUG_OUTPUT
    cout << "||v_0||_2^2 = " << ( wfinfo1s.vWavefunction.transpose() * wfinfo1s.vWavefunction ) << endl;
#endif
    ColumnVector const vWfComb1s =
      ( nIteration > 0 )
      ? ( ( 1.0 - hforbparams1s.rHFSlope ) * vWf1s
          + hforbparams1s.rHFSlope * wfinfo1s.vWavefunction )
      : wfinfo1s.vWavefunction;
    ColumnVector const vWfComb1sScf =
      ConvertVectorToNearestScfBasis(
        wv, basisNsof, anDecNsof, usfbScfBasis, vWfComb1s
        );
    Real const rNC1s =
      ComputeNormalizationCoefficient( usfbScfBasis, vWfComb1sScf );
#ifdef DEBUG_OUTPUT
    cout << "rNC1s = " << rNC1s << endl;
#endif
    vWf1s = vWfComb1s / rNC1s;
    ColumnVector const vWf1sScf = vWfComb1sScf / rNC1s;
#ifdef DEBUG_OUTPUT
    cout << "||v||_2^2 = " << ( vWf1s.transpose() * vWf1s ) << endl;
    cout << "||v_scf||_2^2 = " << ( vWf1sScf.transpose() * vWf1sScf ) << endl;
#endif
#ifdef DEBUG_OUTPUT
    cout << "Computing J operator..." << endl;
#endif
    ColumnVector const vWfZ1sSquare =
      InsertElementToColumnVector(
        ConvertVectorFromScfToStandardBasis(
          wv, basisSof, usfbScfBasis,
          product( vWf1sScf, vWf1sScf )
          ),
        iIndexToZ,
        0.0
        );
    ColumnVector const vY01s1s = mxOpZSlater0 * vWfZ1sSquare;
    ColumnVector const vY01s1sScf =
      ConvertVectorToNearestScfBasis(
        wv, basisZNsof, anDecZNsof, usfbScfBasis, vY01s1s
        );
    Matrix const mxOpZJ1s =
      ComputeOperator(
        wv,
        ProductNsofBlockComputator( wv, usfbScfBasis, vY01s1sScf ),
        basisZNsof,
        anDecZNsof
        );
    mxOpJ1s =
      RemoveZFromMatrix( mxOpZJ1s, iIndexToZ );
#ifdef DEBUG_OUTPUT
    cout << "||JZ1s||_1 = " << ComputeMatrixNorm( mxOpZJ1s, mxnormOne ) << endl;
    cout << "||J1s||_1 = " << ComputeMatrixNorm( mxOpJ1s, mxnormOne ) << endl;
#endif
    if( nIteration == 0 )
      {
      vWf1sHydr = vWf1s;
      rEpsilon1sHydr = rEpsilon1s;
      }
    Real const rConvIndic =
      ComputeMatrixNorm( mxOpF1s - mxOpF1sPrev, nMatrixNormType );
    logger.NotifyConvergenceIndicators( GetColumnVector( rConvIndic ) );
    if( fabs( rConvIndic ) < rConvThreshold )
      {
      fConvergenceObtained = true;
#ifdef DEBUG_OUTPUT
      cout << "Convergence obtained in HF iteration. Iteration stopped."
           << endl;
#endif
      break;
      }
    mxOpF1sPrev = mxOpF1s;
#ifdef DEBUG_OUTPUT
    cout << "Step: " << nIteration << ": Done." << endl;
#endif
    }
  HeliumHFComputationResults oResults;
  oResults.errhf =
    fConvergenceObtained ? errhfNoError : errhfNoConvergence;
  oResults.nErrorInOrbital = -1;
  oResults.rEpsilon1s = rEpsilon1s;
  oResults.rEpsilon1sHydr = rEpsilon1sHydr;
  oResults.vWf1s = vWf1s;
  oResults.vWf1sHydr = vWf1sHydr;
  oResults.mxOpLaplacian = mxOpLaplacian;
  oResults.mxOpR = mxOpR;
  oResults.mxOpInvR = mxOpInvR;
  oResults.mxOpZR = mxOpZR;
  oResults.mxOpZIntZero = mxOpZIntZero;
  oResults.mxOpZIntInf = mxOpZIntInf;
  oResults.mxOpZSlater0 = mxOpZSlater0;
  oResults.mxOpH0L0 = mxOpH0L0;
  oResults.mxOpJ1s = mxOpJ1s;
  return oResults;
  }
