
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of hydrogenic s orbitals

// Implementation file

// Copyright (C) 2004  Tommi Hynlnmaa

#include "DoHydrogenicSComputationWLS.h"
#include "ComputeOperator.h"
#include "LaplacianNsofBlockComputator.h"
#include "RNsofBlockComputator.h"
#include "Triplet.h"
#include "SortWavefunctions.h"

// liboctave headers
#include "EIG.h"

// #define DEBUG_OUTPUT

#ifdef DEBUG_OUTPUT
#include <iostream>
#endif

HydrogenicSComputationResultsWLS DoHydrogenicSComputationWLS(
  WaveletFilters const & wv,
  Array< Integer > anDecSof,
  Basis const & basisNsof,
  Array< Integer > anDecNsof,
  bool fZPresentInBasis,
  UniformBasis const & usfbScfBasis,
  ColumnVector vD1Filter,
  ColumnVector vD2Filter,
  Integer nAtomicNumber,
  Integer nDesiredNodes,
  Real rAntisymmThreshold,
  Real rDerivThreshold,
  Real rSignCheckPosition,
  Real rLengthUnit
  )
  {
#ifdef DEBUG_OUTPUT
  cout << "Computing Laplacian..." << endl;
#endif
  Matrix const mxOpLaplacian =
    ComputeOperator(
      wv,
      LaplacianNsofBlockComputator( wv, vD2Filter ),
      basisNsof,
      anDecNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing R operator..." << endl;
#endif
  Matrix const mxOpR =
    ComputeOperator(
      wv,
      RNsofBlockComputator( wv ),
      basisNsof,
      anDecNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing R^(-1)..." << endl;
#endif
  Integer nInversionInfo = 0;
  Matrix const mxOpInvR = mxOpR.inverse( nInversionInfo );
  if( nInversionInfo == -1 )
    {
    HydrogenicSComputationResultsWLS oResults;
    oResults.wfinfo.fFound = false;
    oResults.errhs = errhsSingularROperator;
    return oResults;
    }
#ifdef DEBUG_OUTPUT
  cout << "Computing Hamiltonian..." << endl;
#endif
  Matrix const mxOpT = 
    ( (-0.5) / ( rLengthUnit * rLengthUnit ) ) * mxOpLaplacian;
  Matrix const mxOpV = ( -nAtomicNumber / rLengthUnit ) * mxOpInvR;
  Matrix const mxOpHamiltonian = mxOpT + mxOpV;
#ifdef DEBUG_OUTPUT
  cout << "Solving eigenproblem..." << endl;
#endif
  EIG eigenproblem( mxOpHamiltonian );
  ComplexColumnVector const cvEigenvalues =
    eigenproblem.eigenvalues();
  ComplexMatrix const cmxEigenvectors =
    eigenproblem.eigenvectors();
#ifdef DEBUG_OUTPUT
  cout << "eigenvalues: " << endl;
  for( Integer i = 0; i < cvEigenvalues.length(); i++ )
    {
    cout << "  " << cvEigenvalues( i ) << endl;
    }
  cout << endl;
  cout << "eigenvectors: " << endl;
  for( Integer i1 = 0; i1 < cmxEigenvectors.rows(); i1++ )
    {
    cout << "  ";
    for( Integer i2 = 0; i2 < cmxEigenvectors.columns(); i2++ )
      {
      cout << cmxEigenvectors( i1, i2 ) << " ";
      }
    cout << endl;
    }
  cout << endl;
#endif
#ifdef DEBUG_OUTPUT
  cout << "Processing results of eigenproblem..." << endl;
#endif
  Triplet< ComplexColumnVector, ComplexMatrix, Array< Integer > > const
    cv_cmx_ai =
    SortWavefunctions( cvEigenvalues, cmxEigenvectors );
  ComplexColumnVector const & cvSortedEigenvalues =
    cv_cmx_ai.first;
  ComplexMatrix const & cmxSortedEigenvectors =
    cv_cmx_ai.second;
  Array< Integer > const & aiSortIndices =
    cv_cmx_ai.third;
#ifdef DEBUG_OUTPUT
  cout << "sorted eigenvalues: " << endl;
  for( Integer i = 0; i < cvSortedEigenvalues.length(); i++ )
    {
    cout << "  " << cvSortedEigenvalues( i ) << endl;
    }
  cout << endl;
  cout << "sorted eigenvectors: " << endl;
  for( Integer i1 = 0; i1 < cmxSortedEigenvectors.rows(); i1++ )
    {
    cout << "  ";
    for( Integer i2 = 0; i2 < cmxSortedEigenvectors.columns(); i2++ )
      {
      cout << cmxSortedEigenvectors( i1, i2 ) << " ";
      }
    cout << endl;
    }
  cout << endl;
#endif
  WavefunctionInfo const wfinfo =
    ChooseWavefunction(
      wv,
      anDecSof,
      fZPresentInBasis,
      basisNsof,
      anDecNsof,
      usfbScfBasis,
      vD1Filter,
      cvSortedEigenvalues,
      cmxSortedEigenvectors,
      nDesiredNodes,
      rAntisymmThreshold,
      rDerivThreshold,
      rSignCheckPosition
      );
  HydrogenicSComputationResultsWLS oResults;
  oResults.wfinfo = wfinfo;
  oResults.errhs =
    ( wfinfo.fFound ) ? errhsNoError : errhsWavefunctionNotFound;
  oResults.cvEigenvalues = cvEigenvalues;
  oResults.cmxEigenvectors = cmxEigenvectors;
  oResults.cvSortedEigenvalues = cvSortedEigenvalues;
  oResults.cmxSortedEigenvectors = cmxSortedEigenvectors;
  oResults.aiSortIndices = aiSortIndices;
  oResults.mxOpT = mxOpT;
  oResults.mxOpV = mxOpV;
  oResults.mxOpHamiltonian = mxOpHamiltonian;
  return oResults;
  }
