
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Computation of hydrogenic orbitals
// with arbitary principal and angular momentum
// quantum numbers (and arbitrary atomic number)

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "DoNeonHFComputation.h"
#include "ComputeOperator.h"
#include "LaplacianNsofBlockComputator.h"
#include "RNsofBlockComputator.h"
#include "ProductNsofBlockComputator.h"
#include "IntegralZeroNsofBlockComputator.h"
#include "IntegralInfNsofBlockComputator.h"
#include "Triplet.h"
#include "SortWavefunctions.h"
#include "ChooseWavefunction.h"
#include "MatrixUtils.h"
#include "VectorUtils.h"
#include "ConvertVectorToNearestScfBasis.h"
#include "ConvertVectorFromScfToStandardBasis.h"
#include "ComputeNormalizationCoefficient.h"
#include "ComputeNormalInnerProduct.h"
// liboctave headers
#include "EIG.h"

#ifdef DEBUG_OUTPUT
#include <iostream>
#endif

#ifdef USE_NAMESPACES
using namespace std;
#endif

static Integer const nDesiredNodes1s = 0;
static Integer const nDesiredNodes2s = 1;
static Integer const nDesiredNodes2p = 0;

NeonHFComputationResults DoNeonHFComputation(
  WaveletFilters const & wv,
  Basis const & basisSof,
  Array< Integer > anDecSof,
  Basis const & basisNsof,
  Array< Integer > anDecNsof,
  Basis const & basisZNsof,
  Array< Integer > anDecZNsof,
  Integer iIndexToZ,
  UniformBasis const & usfbScfBasis,
  MotherIntegral const & motherintegral,
  ColumnVector vD1Filter,
  ColumnVector vD2Filter,
  Integer nAtomicNumber,
  Integer nMaxIterations,
  Real rConvThreshold,
  EnumMatrixNormType nMatrixNormType,
  HFOrbitalComputationParams const & hforbparams1s,
  HFOrbitalComputationParams const & hforbparams2s,
  HFOrbitalComputationParams const & hforbparams2p,
  HFLogger& logger
  )
  {
#ifdef DEBUG_OUTPUT
  cout << "DoNeonHFComputation ENTER" << endl;
#endif
  Integer const nBasis = basisSof.length();
#ifdef DEBUG_OUTPUT
  cout << "Computing Laplacian..." << endl;
#endif
  Matrix const mxOpLaplacian =
    ComputeOperator(
      wv,
      LaplacianNsofBlockComputator( wv, vD2Filter ),
      basisNsof,
      anDecNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing R operator..." << endl;
#endif
  Matrix const mxOpR =
    ComputeOperator(
      wv,
      RNsofBlockComputator( wv ),
      basisNsof,
      anDecNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing R^(-1)..." << endl;
#endif
  Integer nInversionInfo = 0;
  Matrix const mxOpInvR = mxOpR.inverse( nInversionInfo );
  if( nInversionInfo == -1 )
    {
    NeonHFComputationResults oResults;
    oResults.errhf = errhfSingularROperator;
    return oResults;
    }
  Matrix const mxOpInvRSquare = mxOpInvR * mxOpInvR;
#ifdef DEBUG_OUTPUT
  cout << "Computing R operator in Z basis..." << endl;
#endif
  Matrix const mxOpZR =
    ComputeOperator(
      wv,
      RNsofBlockComputator( wv ),
      basisZNsof,
      anDecZNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing integral-from-zero operator in Z basis..." << endl;
#endif
  Matrix const mxOpZIntZero =
    ComputeOperator(
      wv,
      IntegralZeroNsofBlockComputator( wv, motherintegral ),
      basisZNsof,
      anDecZNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing integral-to-infinity operator in Z basis..." 
       << endl;
#endif
  Matrix const mxOpZIntInf =
    ComputeOperator(
      wv,
      IntegralInfNsofBlockComputator( wv, motherintegral ),
      basisZNsof,
      anDecZNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing single-electron Hamiltonians..." << endl;
#endif
  Matrix const mxOpT = (-0.5) * mxOpLaplacian;
  Matrix const mxOpV = ( -nAtomicNumber ) * mxOpInvR;
  Matrix const mxOpC1 = mxOpInvRSquare;
  Matrix const mxOpH0L0 = mxOpT + mxOpV;
  Matrix const mxOpH0L1 = mxOpT + mxOpV + mxOpC1;
#ifdef DEBUG_OUTPUT
  cout << "Computing Slater integral operators..." << endl;
#endif
  Matrix const mxOp10R = 
    RemoveRowFromMatrix( mxOpZR, iIndexToZ );
  Matrix const mxOp01IntInf = 
    RemoveColumnFromMatrix( mxOpZIntInf, iIndexToZ );
  Matrix const mxOp01IntZero =
    RemoveColumnFromMatrix( mxOpZIntZero, iIndexToZ );
  Matrix const mxOp11IntZero =
    RemoveRowFromMatrix( mxOp01IntZero, iIndexToZ );
#ifdef DEBUG_OUTPUT
  cout << mxOpInvR.rows() << ", " << mxOpInvR.columns() << endl;
  cout << mxOp11IntZero.rows() << ", " << mxOp11IntZero.columns()
       << endl;
#endif
  Matrix const mxOp11Slater0FromZero = mxOpInvR * mxOp11IntZero;
  Matrix const mxOp00Slater0FromZero =
    InsertZToMatrix( mxOp11Slater0FromZero, iIndexToZ );
#ifdef DEBUG_OUTPUT
  cout << mxOp01IntInf.rows() << ", " << mxOp01IntInf.columns()
       << endl;
#endif
  Matrix const mxOp01Slater0ToInf = mxOp01IntInf * mxOpInvR;
  Matrix const mxOp00Slater0ToInf =
    InsertColumnToMatrix(
      mxOp01Slater0ToInf,
      ColumnVector( mxOp01Slater0ToInf.rows(), 0.0 ),
      iIndexToZ
      );
  Matrix const mxOpZSlater0 = mxOp00Slater0FromZero + mxOp00Slater0ToInf;
  Matrix const mxOpInvRToThird = mxOpInvR * mxOpInvRSquare;
  Matrix const mxOpRSquare = mxOpR * mxOpR;
  Matrix const mxOpZRSquare = mxOpZR * mxOpZR;
  Matrix const mxOp10R2 = RemoveRowFromMatrix( mxOpZRSquare, iIndexToZ );
#ifdef DEBUG_OUTPUT
  cout << mxOpInvRSquare.rows() << ", " << mxOpInvRSquare.columns()
       << endl;
  cout << mxOp11IntZero.rows() << ", " << mxOp11IntZero.columns()
       << endl;
  cout << mxOpR.rows() << ", " << mxOpR.columns()
       << endl;
#endif
  Matrix const mxOp11Slater1FromZero =
    mxOpInvRSquare * mxOp11IntZero * mxOpR;
  Matrix const mxOp00Slater1FromZero =
    InsertZToMatrix( mxOp11Slater1FromZero, iIndexToZ );
#ifdef DEBUG_OUTPUT
  cout << mxOp10R.rows() << ", " << mxOp10R.columns()
       << endl;
  cout << mxOp01IntInf.rows() << ", " << mxOp01IntInf.columns()
       << endl;
  cout << mxOpInvRSquare.rows() << ", " << mxOpInvRSquare.columns()
       << endl;
#endif
  Matrix const mxOp11Slater1ToInf =
    mxOp10R * mxOp01IntInf * mxOpInvRSquare;
  Matrix const mxOp00Slater1ToInf =
    InsertZToMatrix( mxOp11Slater1ToInf, iIndexToZ );
  Matrix const mxOpZSlater1 =
    mxOp00Slater1FromZero + mxOp00Slater1ToInf;
#ifdef DEBUG_OUTPUT
  cout << mxOpInvRToThird.rows() << ", " << mxOpInvRToThird.columns()
       << endl;
  cout << mxOp11IntZero.rows() << ", " << mxOp11IntZero.columns()
       << endl;
  cout << mxOpRSquare.rows() << ", " << mxOpRSquare.columns()
       << endl;
#endif
  Matrix const mxOp11Slater2FromZero =
    mxOpInvRToThird * mxOp11IntZero * mxOpRSquare;
  Matrix const mxOp00Slater2FromZero =
    InsertZToMatrix( mxOp11Slater2FromZero, iIndexToZ );
#ifdef DEBUG_OUTPUT
  cout << mxOp10R2.rows() << ", " << mxOp10R2.columns()
       << endl;
  cout << mxOp01IntInf.rows() << ", " << mxOp01IntInf.columns()
       << endl;
  cout << mxOpInvRToThird.rows() << ", " << mxOpInvRToThird.columns()
       << endl;
#endif
  Matrix const mxOp11Slater2ToInf =
    mxOp10R2 * mxOp01IntInf * mxOpInvRToThird;
  Matrix const mxOp00Slater2ToInf =
    InsertZToMatrix( mxOp11Slater2ToInf, iIndexToZ );
  Matrix const mxOpZSlater2 =
    mxOp00Slater2FromZero + mxOp00Slater2ToInf;
#ifdef DEBUG_OUTPUT
  cout << "Preparing HF iteration..." << endl;
#endif
  Matrix mxOpF1sPrev( nBasis, nBasis, 0.0 );
  Matrix mxOpF2sPrev( nBasis, nBasis, 0.0 );
  Matrix mxOpF2pPrev( nBasis, nBasis, 0.0 );
  Matrix mxOpF1s( nBasis, nBasis, 0.0 );
  Matrix mxOpF2s( nBasis, nBasis, 0.0 );
  Matrix mxOpF2p( nBasis, nBasis, 0.0 );
  Matrix mxOpJ1s( nBasis, nBasis, 0.0 );
  Matrix mxOpJ2s( nBasis, nBasis, 0.0 );
  Matrix mxOpJ2p( nBasis, nBasis, 0.0 );
  Matrix mxOpK1s( nBasis, nBasis, 0.0 );
  Matrix mxOpK2s( nBasis, nBasis, 0.0 );
  Matrix mxOpK2p( nBasis, nBasis, 0.0 );
  ColumnVector vWf1s( nBasis, 0.0 );
  ColumnVector vWf2s( nBasis, 0.0 );
  ColumnVector vWf2p( nBasis, 0.0 );
  ColumnVector vWf1sHydr( nBasis, 0.0 );
  ColumnVector vWf2sHydr( nBasis, 0.0 );
  ColumnVector vWf2pHydr( nBasis, 0.0 );
  Real rEpsilon1s = 0.0;
  Real rEpsilon2s = 0.0;
  Real rEpsilon2p = 0.0;
  Real rEpsilon1sHydr = 0.0;
  Real rEpsilon2sHydr = 0.0;
  Real rEpsilon2pHydr = 0.0;
  bool fConvergenceObtained = false;
#ifdef DEBUG_OUTPUT
  cout << "Hartree-Fock iteration..." << endl;
#endif
  for( Integer nIteration = 0; 
       nIteration < nMaxIterations; 
       nIteration++ )
    {
    logger.NotifyIterationStep( nIteration );
#ifdef DEBUG_OUTPUT
    cout << "Computing Fock operators" << endl;
#endif
    mxOpF1s = mxOpH0L0 + mxOpJ1s - mxOpK1s;
    mxOpF2s = mxOpH0L0 + mxOpJ2s - mxOpK2s;
    mxOpF2p = mxOpH0L1 + mxOpJ2p - mxOpK2p;
#ifdef DEBUG_OUTPUT
    cout << "Solving eigenproblem for 1s..." << endl;
#endif
    EIG eigenproblem1s( mxOpF1s );
    ComplexColumnVector const cvEigenvalues1s =
      eigenproblem1s.eigenvalues();
    ComplexMatrix const cmxEigenvectors1s =
      eigenproblem1s.eigenvectors();
    Triplet< ComplexColumnVector, ComplexMatrix, Array< Integer > > const
      cv_cmx_ai1s =
      SortWavefunctions( cvEigenvalues1s, cmxEigenvectors1s );
    ComplexColumnVector const & cvSortedEigenvalues1s =
      cv_cmx_ai1s.first;
    ComplexMatrix const & cmxSortedEigenvectors1s =
      cv_cmx_ai1s.second;
    WavefunctionInfo const wfinfo1s =
      ChooseWavefunction(
        wv,
        anDecSof,
        false,
        basisNsof,
        anDecNsof,
        usfbScfBasis,
        vD1Filter,
        cvSortedEigenvalues1s,
        cmxSortedEigenvectors1s,
        nDesiredNodes1s,
        hforbparams1s.rAntisymmThreshold,
        hforbparams1s.rDerivThreshold,
        hforbparams1s.rSignCheckPosition
        );
    if( ! wfinfo1s.fFound )
      {
#ifdef DEBUG_OUTPUT
      cout << "1s Wavefunction not found" << endl;
#endif
      NeonHFComputationResults oResults;
      oResults.errhf = errhfWavefunctionNotFound;
      oResults.nErrorAtPQN = 1;
      oResults.nErrorAtAMQN = 0;
      oResults.mxOpLaplacian = mxOpLaplacian;
      oResults.mxOpR = mxOpR;
      oResults.mxOpInvR = mxOpInvR;
      oResults.mxOpZR = mxOpZR;
      oResults.mxOpZIntZero = mxOpZIntZero;
      oResults.mxOpZIntInf = mxOpZIntInf;
      oResults.mxOpZSlater0 = mxOpZSlater0;
      oResults.mxOpZSlater1 = mxOpZSlater1;
      oResults.mxOpZSlater2 = mxOpZSlater2;
      oResults.mxOpH0L0 = mxOpH0L0;
      oResults.mxOpH0L1 = mxOpH0L1;
      oResults.mxOpJ1s = mxOpJ1s;
      oResults.mxOpJ2s = mxOpJ2s;
      oResults.mxOpJ2p = mxOpJ2p;
      oResults.mxOpK1s = mxOpK1s;
      oResults.mxOpK2s = mxOpK2s;
      oResults.mxOpK2p = mxOpK2p;
      return oResults;
      }
    rEpsilon1s = wfinfo1s.rEpsilon;
    logger.NotifyOrbitalInfo( 0, 1, 0, rEpsilon1s, wfinfo1s.anRejectionReasons );
#ifdef DEBUG_OUTPUT
    cout << "Solving eigenproblem for 2s..." << endl;
#endif
    EIG eigenproblem2s( mxOpF2s );
    ComplexColumnVector const cvEigenvalues2s =
      eigenproblem2s.eigenvalues();
    ComplexMatrix const cmxEigenvectors2s =
      eigenproblem2s.eigenvectors();
    Triplet< ComplexColumnVector, ComplexMatrix, Array< Integer > > const
      cv_cmx_ai2s =
      SortWavefunctions( cvEigenvalues2s, cmxEigenvectors2s );
    ComplexColumnVector const & cvSortedEigenvalues2s =
      cv_cmx_ai2s.first;
    ComplexMatrix const & cmxSortedEigenvectors2s =
      cv_cmx_ai2s.second;
    WavefunctionInfo const wfinfo2s =
      ChooseWavefunction(
        wv,
        anDecSof,
        false,
        basisNsof,
        anDecNsof,
        usfbScfBasis,
        vD1Filter,
        cvSortedEigenvalues2s,
        cmxSortedEigenvectors2s,
        nDesiredNodes2s,
        hforbparams2s.rAntisymmThreshold,
        hforbparams2s.rDerivThreshold,
        hforbparams2s.rSignCheckPosition
        );
    if( ! wfinfo2s.fFound )
      {
#ifdef DEBUG_OUTPUT
      cout << "2s Wavefunction not found" << endl;
#endif
      NeonHFComputationResults oResults;
      oResults.errhf = errhfWavefunctionNotFound;
      oResults.nErrorAtPQN = 2;
      oResults.nErrorAtAMQN = 0;
      oResults.mxOpLaplacian = mxOpLaplacian;
      oResults.mxOpR = mxOpR;
      oResults.mxOpInvR = mxOpInvR;
      oResults.mxOpZR = mxOpZR;
      oResults.mxOpZIntZero = mxOpZIntZero;
      oResults.mxOpZIntInf = mxOpZIntInf;
      oResults.mxOpZSlater0 = mxOpZSlater0;
      oResults.mxOpZSlater1 = mxOpZSlater1;
      oResults.mxOpZSlater2 = mxOpZSlater2;
      oResults.mxOpH0L0 = mxOpH0L0;
      oResults.mxOpH0L1 = mxOpH0L1;
      oResults.mxOpJ1s = mxOpJ1s;
      oResults.mxOpJ2s = mxOpJ2s;
      oResults.mxOpJ2p = mxOpJ2p;
      oResults.mxOpK1s = mxOpK1s;
      oResults.mxOpK2s = mxOpK2s;
      oResults.mxOpK2p = mxOpK2p;
      return oResults;
      }
    rEpsilon2s = wfinfo2s.rEpsilon;
    logger.NotifyOrbitalInfo( 1, 2, 0, rEpsilon1s, wfinfo1s.anRejectionReasons );
#ifdef DEBUG_OUTPUT
    cout << "Solving eigenproblem for 2p..." << endl;
#endif
    EIG eigenproblem2p( mxOpF2p );
    ComplexColumnVector const cvEigenvalues2p =
      eigenproblem2p.eigenvalues();
    ComplexMatrix const cmxEigenvectors2p =
      eigenproblem2p.eigenvectors();
    Triplet< ComplexColumnVector, ComplexMatrix, Array< Integer > > const
      cv_cmx_ai2p =
      SortWavefunctions( cvEigenvalues2p, cmxEigenvectors2p );
    ComplexColumnVector const & cvSortedEigenvalues2p =
      cv_cmx_ai2p.first;
    ComplexMatrix const & cmxSortedEigenvectors2p =
      cv_cmx_ai2p.second;
    WavefunctionInfo const wfinfo2p =
      ChooseWavefunction(
        wv,
        anDecSof,
        false,
        basisNsof,
        anDecNsof,
        usfbScfBasis,
        vD1Filter,
        cvSortedEigenvalues2p,
        cmxSortedEigenvectors2p,
        nDesiredNodes2p,
        hforbparams2p.rAntisymmThreshold,
        hforbparams2p.rDerivThreshold,
        hforbparams2p.rSignCheckPosition
        );
    if( ! wfinfo2p.fFound )
      {
#ifdef DEBUG_OUTPUT
      cout << "2p Wavefunction not found" << endl;
#endif
      NeonHFComputationResults oResults;
      oResults.errhf = errhfWavefunctionNotFound;
      oResults.nErrorAtPQN = 2;
      oResults.nErrorAtAMQN = 1;
      oResults.mxOpLaplacian = mxOpLaplacian;
      oResults.mxOpR = mxOpR;
      oResults.mxOpInvR = mxOpInvR;
      oResults.mxOpZR = mxOpZR;
      oResults.mxOpZIntZero = mxOpZIntZero;
      oResults.mxOpZIntInf = mxOpZIntInf;
      oResults.mxOpZSlater0 = mxOpZSlater0;
      oResults.mxOpZSlater1 = mxOpZSlater1;
      oResults.mxOpZSlater2 = mxOpZSlater2;
      oResults.mxOpH0L0 = mxOpH0L0;
      oResults.mxOpH0L1 = mxOpH0L1;
      oResults.mxOpJ1s = mxOpJ1s;
      oResults.mxOpJ2s = mxOpJ2s;
      oResults.mxOpJ2p = mxOpJ2p;
      oResults.mxOpK1s = mxOpK1s;
      oResults.mxOpK2s = mxOpK2s;
      oResults.mxOpK2p = mxOpK2p;
      return oResults;
      }
    rEpsilon2p = wfinfo2p.rEpsilon;
    logger.NotifyOrbitalInfo( 2, 2, 1, rEpsilon1s, wfinfo1s.anRejectionReasons );
#ifdef DEBUG_OUTPUT
    cout << "Computing final wavefunctions for current step..." 
         << endl;
#endif
    ColumnVector const vWfComb1s =
      ( nIteration > 0 )
      ? ( ( 1.0 - hforbparams1s.rHFSlope ) * vWf1s
          + hforbparams1s.rHFSlope * wfinfo1s.vWavefunction )
      : wfinfo1s.vWavefunction;
    ColumnVector const vWfComb2s =
      ( nIteration > 0 )
      ? ( ( 1.0 - hforbparams2s.rHFSlope ) * vWf2s
          + hforbparams2s.rHFSlope * wfinfo2s.vWavefunction )
      : wfinfo2s.vWavefunction;
    ColumnVector const vWfComb2p =
      ( nIteration > 0 )
      ? ( ( 1.0 - hforbparams2p.rHFSlope ) * vWf2p
          + hforbparams2p.rHFSlope * wfinfo2p.vWavefunction )
      : wfinfo2p.vWavefunction;
    ColumnVector const vWfComb1sScf =
      ConvertVectorToNearestScfBasis(
        wv, basisNsof, anDecNsof, usfbScfBasis, vWfComb1s
        );
    ColumnVector const vWfComb2sScf =
      ConvertVectorToNearestScfBasis(
        wv, basisNsof, anDecNsof, usfbScfBasis, vWfComb2s
        );
    ColumnVector const vWfComb2pScf =
      ConvertVectorToNearestScfBasis(
        wv, basisNsof, anDecNsof, usfbScfBasis, vWfComb2p
        );
    Real const rNC1s =
      ComputeNormalizationCoefficient( usfbScfBasis, vWfComb1sScf );
    Real const rNC2s =
      ComputeNormalizationCoefficient( usfbScfBasis, vWfComb2sScf );
    Real const rNC2p =
      ComputeNormalizationCoefficient( usfbScfBasis, vWfComb2pScf );
    vWf1s = vWfComb1s / rNC1s;
    vWf2s = vWfComb2s / rNC2s;
    vWf2p = vWfComb2p / rNC2p;
    ColumnVector const vWf1sScf = vWfComb1sScf / rNC1s;
    ColumnVector const vWf2sScf = vWfComb2sScf / rNC2s;
    ColumnVector const vWf2pScf = vWfComb2pScf / rNC2p;
#ifdef DEBUG_OUTPUT
    Real const rIP1s1s =
      ComputeNormalInnerProduct( usfbScfBasis, vWf1sScf, vWf1sScf );
    Real const rIP2s2s =
      ComputeNormalInnerProduct( usfbScfBasis, vWf2sScf, vWf2sScf );
    Real const rIP2p2p =
      ComputeNormalInnerProduct( usfbScfBasis, vWf2pScf, vWf2pScf );
    Real const rIP1s2s =
      ComputeNormalInnerProduct( usfbScfBasis, vWf1sScf, vWf2sScf );
    cout << "Step: " << nIteration 
         << ": < 1s | 1s > = " << rIP1s1s
         << endl;
    cout << "Step: " << nIteration 
         << ": < 2s | 2s > = " << rIP2s2s
         << endl;
    cout << "Step: " << nIteration 
         << ": < 2p | 2p > = " << rIP2p2p
         << endl;
    cout << "Step: " << nIteration 
         << ": < 1s | 2s > = " << rIP1s2s
         << endl;
#endif
#ifdef DEBUG_OUTPUT
    cout << "Computing J and K operators..." << endl;
#endif
    ColumnVector const vWfZ1sSquare =
      InsertElementToColumnVector(
        ConvertVectorFromScfToStandardBasis(
          wv, basisSof, usfbScfBasis,
          product( vWf1sScf, vWf1sScf )
          ),
        iIndexToZ,
        0.0
        );
    ColumnVector const vY01s1s = mxOpZSlater0 * vWfZ1sSquare;
    ColumnVector const vY01s1sScf =
      ConvertVectorToNearestScfBasis(
        wv, basisZNsof, anDecZNsof, usfbScfBasis, vY01s1s
        );
    Matrix const mxOpZJ01s =
      ComputeOperator(
        wv,
        ProductNsofBlockComputator( wv, usfbScfBasis, vY01s1sScf ),
        basisZNsof,
        anDecZNsof
        );
    Matrix const mxOpJ01s =
      RemoveZFromMatrix( mxOpZJ01s, iIndexToZ );
    ColumnVector const vWfZ2sSquare =
      InsertElementToColumnVector(
        ConvertVectorFromScfToStandardBasis(
          wv, basisSof, usfbScfBasis,
          product( vWf2sScf, vWf2sScf )
          ),
        iIndexToZ,
        0.0
        );
    ColumnVector const vY02s2s = mxOpZSlater0 * vWfZ2sSquare;
    ColumnVector const vY02s2sScf =
      ConvertVectorToNearestScfBasis(
        wv, basisZNsof, anDecZNsof, usfbScfBasis, vY02s2s
        );
    Matrix const mxOpZJ02s =
      ComputeOperator(
        wv,
        ProductNsofBlockComputator( wv, usfbScfBasis, vY02s2sScf ),
        basisZNsof,
        anDecZNsof
        );
    Matrix const mxOpJ02s =
      RemoveZFromMatrix( mxOpZJ02s, iIndexToZ );
    ColumnVector const vWfZ2pSquare =
      InsertElementToColumnVector(
        ConvertVectorFromScfToStandardBasis(
          wv, basisSof, usfbScfBasis,
          product( vWf2pScf, vWf2pScf )
          ),
        iIndexToZ,
        0.0
        );
    ColumnVector const vY02p2p = mxOpZSlater0 * vWfZ2pSquare;
    ColumnVector const vY02p2pScf =
      ConvertVectorToNearestScfBasis(
        wv, basisZNsof, anDecZNsof, usfbScfBasis, vY02p2p
        );
    Matrix const mxOpZJ02p =
      ComputeOperator(
        wv,
        ProductNsofBlockComputator( wv, usfbScfBasis, vY02p2pScf ),
        basisZNsof,
        anDecZNsof
        );
    Matrix const mxOpJ02p =
      RemoveZFromMatrix( mxOpZJ02p, iIndexToZ );
    ColumnVector const vY22p2p = mxOpZSlater2 * vWfZ2pSquare;
    ColumnVector const vY22p2pScf =
      ConvertVectorToNearestScfBasis(
        wv, basisZNsof, anDecZNsof, usfbScfBasis, vY22p2p
        );
    Matrix const mxOpZJ22p =
      ComputeOperator(
        wv,
        ProductNsofBlockComputator( wv, usfbScfBasis, vY22p2pScf ),
        basisZNsof,
        anDecZNsof
        );
    Matrix const mxOpJ22p =
      RemoveZFromMatrix( mxOpZJ22p, iIndexToZ );
    Matrix const mxOpZProd1s =
      ComputeOperator(
        wv,
        ProductNsofBlockComputator( wv, usfbScfBasis, vWf1sScf ),
        basisZNsof,
        anDecZNsof
        );
    Matrix const mxOpZProd2s =
      ComputeOperator(
        wv,
        ProductNsofBlockComputator( wv, usfbScfBasis, vWf2sScf ),
        basisZNsof,
        anDecZNsof
        );
    Matrix const mxOpZProd2p =
      ComputeOperator(
        wv,
        ProductNsofBlockComputator( wv, usfbScfBasis, vWf2pScf ),
        basisZNsof,
        anDecZNsof
        );
    Matrix const mxOpZK01s = mxOpZProd1s * mxOpZSlater0 * mxOpZProd1s;
    Matrix const mxOpZK11s = mxOpZProd1s * mxOpZSlater1 * mxOpZProd1s;
    Matrix const mxOpZK02s = mxOpZProd2s * mxOpZSlater0 * mxOpZProd2s;
    Matrix const mxOpZK12s = mxOpZProd2s * mxOpZSlater1 * mxOpZProd2s;
    Matrix const mxOpZK12p = mxOpZProd2p * mxOpZSlater1 * mxOpZProd2p;
    Matrix const mxOpK01s = RemoveZFromMatrix( mxOpZK01s, iIndexToZ );
    Matrix const mxOpK11s = RemoveZFromMatrix( mxOpZK11s, iIndexToZ );
    Matrix const mxOpK02s = RemoveZFromMatrix( mxOpZK02s, iIndexToZ );
    Matrix const mxOpK12s = RemoveZFromMatrix( mxOpZK12s, iIndexToZ );
    Matrix const mxOpK12p = RemoveZFromMatrix( mxOpZK12p, iIndexToZ );
    mxOpJ1s = mxOpJ01s + 2 * mxOpJ02s + 6 * mxOpJ02p;
    mxOpK1s = mxOpK02s + mxOpK12p;
    mxOpJ2s = 2 * mxOpJ01s + mxOpJ02s + 6 * mxOpJ02p;
    mxOpK2s = mxOpK01s + mxOpK12p;
    mxOpJ2p = 2 * mxOpJ01s + 2 * mxOpJ02s + 5 * mxOpJ02p
      - ( 2.0 / 5.0 ) * mxOpJ22p;
    mxOpK2p = ( 1.0 / 3.0 ) * mxOpK11s + ( 1.0 / 3.0 ) * mxOpK12s;
    if( nIteration == 0 )
      {
      vWf1sHydr = vWf1s;
      vWf2sHydr = vWf2s;
      vWf2pHydr = vWf2p;
      rEpsilon1sHydr = rEpsilon1s;
      rEpsilon2sHydr = rEpsilon2s;
      rEpsilon2pHydr = rEpsilon2p;
      }
    ColumnVector vConvIndic( 3 ); // 3 orbitals in Neon
    vConvIndic( 0 ) =
      ComputeMatrixNorm( mxOpF1s - mxOpF1sPrev, nMatrixNormType );
    vConvIndic( 1 ) =
      ComputeMatrixNorm( mxOpF2s - mxOpF2sPrev, nMatrixNormType );
    vConvIndic( 2 ) =
      ComputeMatrixNorm( mxOpF2p - mxOpF2pPrev, nMatrixNormType );
    logger.NotifyConvergenceIndicators( vConvIndic );
    if( GetColumnVectorInfinityNorm( vConvIndic ) < rConvThreshold )
      {
      fConvergenceObtained = true;
#ifdef DEBUG_OUTPUT
      cout << "Convergence obtained in HF iteration. Iteration stopped."
           << endl;
#endif
      break;
      }
    mxOpF1sPrev = mxOpF1s;
    mxOpF2sPrev = mxOpF2s;
    mxOpF2pPrev = mxOpF2p;
#ifdef DEBUG_OUTPUT
    cout << "Step: " << nIteration << ": Done." << endl;
#endif
    }
  NeonHFComputationResults oResults;
  oResults.errhf =
    fConvergenceObtained ? errhfNoError : errhfNoConvergence;
  oResults.nErrorAtPQN = -1;
  oResults.nErrorAtAMQN = -1;
  oResults.rEpsilon1s = rEpsilon1s;
  oResults.rEpsilon2s = rEpsilon2s;
  oResults.rEpsilon2p = rEpsilon2p;
  oResults.rEpsilon1sHydr = rEpsilon1sHydr;
  oResults.rEpsilon2sHydr = rEpsilon2sHydr;
  oResults.rEpsilon2pHydr = rEpsilon2pHydr;
  oResults.vWf1s = vWf1s;
  oResults.vWf2s = vWf2s;
  oResults.vWf2p = vWf2p;
  oResults.vWf1sHydr = vWf1sHydr;
  oResults.vWf2sHydr = vWf2sHydr;
  oResults.vWf2pHydr = vWf2pHydr;
  oResults.mxOpLaplacian = mxOpLaplacian;
  oResults.mxOpR = mxOpR;
  oResults.mxOpInvR = mxOpInvR;
  oResults.mxOpZR = mxOpZR;
  oResults.mxOpZIntZero = mxOpZIntZero;
  oResults.mxOpZIntInf = mxOpZIntInf;
  oResults.mxOpZSlater0 = mxOpZSlater0;
  oResults.mxOpZSlater1 = mxOpZSlater1;
  oResults.mxOpZSlater2 = mxOpZSlater2;
  oResults.mxOpH0L0 = mxOpH0L0;
  oResults.mxOpH0L1 = mxOpH0L1;
  oResults.mxOpJ1s = mxOpJ1s;
  oResults.mxOpJ2s = mxOpJ2s;
  oResults.mxOpJ2p = mxOpJ2p;
  oResults.mxOpK1s = mxOpK1s;
  oResults.mxOpK2s = mxOpK2s;
  oResults.mxOpK2p = mxOpK2p;
  return oResults;
  }
