
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Hartree-Fock computation where nondiagonal Lagrange
// multipliers are neglected.

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "DoNondiagNeglHFComputation.h"
#include "ComputeOperator.h"
#include "LaplacianNsofBlockComputator.h"
#include "RNsofBlockComputator.h"
#include "ProductNsofBlockComputator.h"
#include "IntegralZeroNsofBlockComputator.h"
#include "IntegralInfNsofBlockComputator.h"
#include "Triplet.h"
#include "SortWavefunctions.h"
#include "ChooseWavefunction.h"
#include "MatrixUtils.h"
#include "VectorUtils.h"
#include "ConvertVectorToNearestScfBasis.h"
#include "ConvertVectorFromScfToStandardBasis.h"
#include "ComputeNormalizationCoefficient.h"
#include "ComputeNormalInnerProduct.h"
// liboctave headers
#include "EIG.h"

#ifdef DEBUG_OUTPUT
#include <iostream>
#endif

NondiagNeglHFComputationResults DoNondiagNeglHFComputation(
  HFAtomicData const & hfatomicdata,
  HFBasisData const & hfbasisdata,
  WaveletFilters const & wv,
  MotherIntegral const & motherintegral,
  ColumnVector vD1Filter,
  ColumnVector vD2Filter,
  Integer nMaxIterations,
  Real rConvThreshold,
  EnumMatrixNormType nMatrixNormType,
  Array< HFOrbitalComputationParams > aHFOrbParams,
  HFLogger& logger
  )
  {
#ifdef DEBUG_OUTPUT
  cout << "DoNondiagNeglHFComputation ENTER" << endl;
#endif
  Integer const nBasis = hfbasisdata.basisSof.length();
  Integer const nOrbitals = hfatomicdata.nOrbitals;
  Integer const nSlaterExps =  hfatomicdata.nSlaterExps;
  if( hfatomicdata.anPQN.length() != hfatomicdata.nOrbitals )
    {
    OwnError( "DoNondiagNeglHFComputation: error in argument hfatomicdata.anPQN" );
    NondiagNeglHFComputationResults oResults;
    oResults.errhf = errhfOther;
    return oResults;
    }
  if( hfatomicdata.anAMQN.length() != hfatomicdata.nOrbitals )
    {
    OwnError( "DoNondiagNeglHFComputation: error in argument hfatomicdata.anAMQN" );
    NondiagNeglHFComputationResults oResults;
    oResults.errhf = errhfOther;
    return oResults;
    }
  if( hfatomicdata.nOrbitals < 1 )
    {
    OwnError( "DoNondiagNeglHFComputation: invalid number of orbitals" );
    NondiagNeglHFComputationResults oResults;
    oResults.errhf = errhfOther;
    return oResults;
    }
  if( 
    hfatomicdata.avCD.rows() != hfatomicdata.nOrbitals ||
    hfatomicdata.avCD.columns() != hfatomicdata.nOrbitals 
    )
    {
    OwnError( "DoNondiagNeglHFComputation: error in argument hfatomicdata.avCD" );
    NondiagNeglHFComputationResults oResults;
    oResults.errhf = errhfOther;
    return oResults;
    }
  if( 
    hfatomicdata.avCE.rows() != hfatomicdata.nOrbitals ||
    hfatomicdata.avCE.columns() != hfatomicdata.nOrbitals 
    )
    {
    OwnError( "DoNondiagNeglHFComputation: error in argument hfatomicdata.avCE" );
    NondiagNeglHFComputationResults oResults;
    oResults.errhf = errhfOther;
    return oResults;
    }
  if( hfatomicdata.anJCounts.length() != hfatomicdata.nOrbitals )
    {
    OwnError( "DoNondiagNeglHFComputation: error in argument hfatomicdata.anJCounts" );
    NondiagNeglHFComputationResults oResults;
    oResults.errhf = errhfOther;
    return oResults;
    }
  if( hfatomicdata.anKCounts.length() != hfatomicdata.nOrbitals )
    {
    OwnError( "DoNondiagNeglHFComputation: error in argument hfatomicdata.anJCounts" );
    NondiagNeglHFComputationResults oResults;
    oResults.errhf = errhfOther;
    return oResults;
    }
  if( aHFOrbParams.length() != hfatomicdata.nOrbitals )
    {
    OwnError( "DoNondiagNeglHFComputation: error in argument aHFOrbParams" );
    NondiagNeglHFComputationResults oResults;
    oResults.errhf = errhfOther;
    return oResults;
    }
#ifdef DEBUG_OUTPUT
  cout << "Computing Laplacian..." << endl;
#endif
  Matrix const mxOpLaplacian =
    ComputeOperator(
      wv,
      LaplacianNsofBlockComputator( wv, vD2Filter ),
      hfbasisdata.basisNsof,
      hfbasisdata.anDecNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing R operator..." << endl;
#endif
  Matrix const mxOpR =
    ComputeOperator(
      wv,
      RNsofBlockComputator( wv ),
      hfbasisdata.basisNsof,
      hfbasisdata.anDecNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing R^(-1)..." << endl;
#endif
  Integer nInversionInfo = 0;
  Matrix const mxOpInvR = mxOpR.inverse( nInversionInfo );
  if( nInversionInfo == -1 )
    {
    NondiagNeglHFComputationResults oResults;
    oResults.errhf = errhfSingularROperator;
    return oResults;
    }
  Matrix const mxOpInvRSquare = mxOpInvR * mxOpInvR;
#ifdef DEBUG_OUTPUT
  cout << "Computing R operator in Z basis..." << endl;
#endif
  Matrix const mxOpZR =
    ComputeOperator(
      wv,
      RNsofBlockComputator( wv ),
      hfbasisdata.basisZNsof,
      hfbasisdata.anDecZNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing integral-from-zero operator in Z basis..." << endl;
#endif
  Matrix const mxOpZIntZero =
    ComputeOperator(
      wv,
      IntegralZeroNsofBlockComputator( wv, motherintegral ),
      hfbasisdata.basisZNsof,
      hfbasisdata.anDecZNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing integral-to-infinity operator in Z basis..." 
       << endl;
#endif
  Matrix const mxOpZIntInf =
    ComputeOperator(
      wv,
      IntegralInfNsofBlockComputator( wv, motherintegral ),
      hfbasisdata.basisZNsof,
      hfbasisdata.anDecZNsof
      );
#ifdef DEBUG_OUTPUT
  cout << "Computing single-electron Hamiltonians..." << endl;
#endif
  Matrix const mxOpT = (-0.5) * mxOpLaplacian;
  Matrix const mxOpV = ( -hfatomicdata.nAtomicNumber ) * mxOpInvR;
  Matrix const mxOpC = mxOpInvRSquare;
  Array< Matrix > amxOpH0( nOrbitals );
  for( Integer iOrbital = 0; iOrbital < nOrbitals; iOrbital++ )
    {
    Integer const nAMQN = hfatomicdata.anAMQN( iOrbital );
    amxOpH0( iOrbital ) =
      mxOpT + mxOpV 
      + ( 0.5 * nAMQN * ( nAMQN + 1 ) ) * mxOpC;
    }
#ifdef DEBUG_OUTPUT
  cout << "Computing Slater integral operators..." << endl;
#endif
  Array< Matrix > amxOpsZSlater( nSlaterExps );
  if( nSlaterExps > 0 )
    {
    Array< Matrix > amxOpsZR( nSlaterExps );
    Array< Matrix > amxOpsInvR( nSlaterExps );
    amxOpsZR( 0 ) = mxOpZR;
    for( Integer i = 1; i < nSlaterExps; i++ )
      {
      amxOpsZR( i ) = mxOpZR * amxOpsZR( i - 1 );
      }
    amxOpsInvR( 0 ) = mxOpInvR * mxOpInvR;
    for( Integer i = 1; i < nSlaterExps; i++ )
      {
      amxOpsInvR( i ) = mxOpInvR * amxOpsInvR( i - 1 );
      }
    // exponent k = 0 handled specially
    Matrix const mxOp01IntInf = 
      RemoveColumnFromMatrix( mxOpZIntInf, hfbasisdata.iIndexToZ );
    Matrix const mxOp01IntZero =
      RemoveColumnFromMatrix( mxOpZIntZero, hfbasisdata.iIndexToZ );
    Matrix const mxOp11IntZero =
      RemoveRowFromMatrix( mxOp01IntZero, hfbasisdata.iIndexToZ );
    Matrix const mxOp11Slater0FromZero = mxOpInvR * mxOp11IntZero;
    Matrix const mxOp00Slater0FromZero =
      InsertZToMatrix( mxOp11Slater0FromZero, hfbasisdata.iIndexToZ );
    Matrix const mxOp01Slater0ToInf = mxOp01IntInf * mxOpInvR;
    Matrix const mxOp00Slater0ToInf =
      InsertColumnToMatrix(
        mxOp01Slater0ToInf,
        ColumnVector( mxOp01Slater0ToInf.rows(), 0.0 ),
        hfbasisdata.iIndexToZ
        );
    amxOpsZSlater( 0 ) =
      mxOp00Slater0FromZero + mxOp00Slater0ToInf;
    for( Integer i = 1; i < nSlaterExps; i++ )
      {
      Matrix const mxOpZRCurr = amxOpsZR( i - 1 );
      Matrix const mxOpInvRCurr = amxOpsInvR( i - 1 );
      Matrix const mxOp10RCurr =
        RemoveRowFromMatrix( mxOpZRCurr, hfbasisdata.iIndexToZ );
      Matrix const mxOpRCurr =
        RemoveColumnFromMatrix( mxOp10RCurr, hfbasisdata.iIndexToZ );
      Matrix const mxOp11FromZeroCurr =
        mxOpInvRCurr * mxOp11IntZero * mxOpRCurr;
      Matrix const mxOp00FromZeroCurr =
        InsertZToMatrix( mxOp11FromZeroCurr, hfbasisdata.iIndexToZ );
      Matrix const mxOp11ToInfCurr =
        mxOp10RCurr * mxOp01IntInf * mxOpInvRCurr;
      Matrix const mxOp00ToInfCurr =
        InsertZToMatrix( mxOp11ToInfCurr, hfbasisdata.iIndexToZ );
      amxOpsZSlater( i ) = mxOp00FromZeroCurr + mxOp00ToInfCurr;
      }
    }
#ifdef DEBUG_OUTPUT
  cout << "Preparing HF iteration..." << endl;
#endif
  Array< Matrix > amxOpFPrev( nOrbitals, Matrix( nBasis, nBasis, 0.0 ) );
  Array< Matrix > amxOpF( nOrbitals, Matrix( nBasis, nBasis, 0.0 ) );
  Array< Matrix > amxOpJ( nOrbitals, Matrix( nBasis, nBasis, 0.0 ) );
  Array< Matrix > amxOpK( nOrbitals, Matrix( nBasis, nBasis, 0.0 ) );
  Matrix mxWf( nBasis, nOrbitals, 0.0 );
  Matrix mxWfHydr( nBasis, nOrbitals, 0.0 );
  ColumnVector vEpsilons( nOrbitals, 0.0 );
  ColumnVector vEpsilonsHydr( nOrbitals, 0.0 );
  NondiagNeglHFComputationResults oResults;
  oResults.errhf = errhfOther;
  oResults.nErrorInOrbital = -1;
  oResults.mxOpLaplacian = mxOpLaplacian;
  oResults.mxOpR = mxOpR;
  oResults.mxOpInvR = mxOpInvR;
  oResults.mxOpZR = mxOpZR;
  oResults.mxOpZIntZero = mxOpZIntZero;
  oResults.mxOpZIntInf = mxOpZIntInf;
  oResults.amxOpsZSlater = amxOpsZSlater;
  bool fConvergenceObtained = false;
#ifdef DEBUG_OUTPUT
  cout << "Hartree-Fock iteration..." << endl;
#endif
  for( Integer nIteration = 0; 
       nIteration < nMaxIterations; 
       nIteration++ )
    {
    logger.NotifyIterationStep( nIteration );
    Array< Array< Matrix > > aamxOpJ( nOrbitals );
    Array< Array< Matrix > > aamxOpK( nOrbitals );
    for( Integer iOrbital = 0;
         iOrbital < nOrbitals;
         iOrbital++ )
      {
#ifdef DEBUG_OUTPUT
      cout << "  Orbital " << iOrbital << endl;
#endif
      Integer const nPQN = hfatomicdata.anPQN( iOrbital );
      Integer const nAMQN = hfatomicdata.anAMQN( iOrbital );
      amxOpF( iOrbital ) =
        amxOpH0( iOrbital ) + amxOpJ( iOrbital ) - amxOpK( iOrbital );
      EIG eigenproblem( amxOpF( iOrbital ) );
      ComplexColumnVector const cvEigenvalues =
        eigenproblem.eigenvalues();
      ComplexMatrix const cmxEigenvectors =
        eigenproblem.eigenvectors();
      Triplet< ComplexColumnVector, ComplexMatrix, Array< Integer > >
        const
        cv_cmx_ai =
        SortWavefunctions( cvEigenvalues, cmxEigenvectors );
      ComplexColumnVector const & cvSortedEigenvalues =
        cv_cmx_ai.first;
      ComplexMatrix const & cmxSortedEigenvectors =
        cv_cmx_ai.second;
      WavefunctionInfo const wfinfo =
        ChooseWavefunction(
          wv,
          hfbasisdata.anDecSof,
          false,
          hfbasisdata.basisNsof,
          hfbasisdata.anDecNsof,
          hfbasisdata.usfbScfBasis,
          vD1Filter,
          cvSortedEigenvalues,
          cmxSortedEigenvectors,
          nPQN - nAMQN - 1, // check this
          aHFOrbParams( iOrbital ).rAntisymmThreshold,
          aHFOrbParams( iOrbital ).rDerivThreshold,
          aHFOrbParams( iOrbital ).rSignCheckPosition
          );
      if( ! wfinfo.fFound )
        {
#ifdef DEBUG_OUTPUT
        cout << "Wavefunction for orbital " << iOrbital 
             << "not found" << endl;
#endif
        oResults.errhf = errhfWavefunctionNotFound;
        oResults.nErrorInOrbital = iOrbital;
        return oResults;
        }
      vEpsilons( iOrbital ) = wfinfo.rEpsilon;
      Real const rHFSlope = aHFOrbParams( iOrbital ).rHFSlope;
      ColumnVector const vWfComb =
        ( nIteration > 0 )
        ? ( ( 1.0 - rHFSlope ) * mxWf.column( iOrbital )
            + rHFSlope * wfinfo.vWavefunction )
        : wfinfo.vWavefunction;
      ColumnVector const vWfCombScf =
        ConvertVectorToNearestScfBasis(
          wv, hfbasisdata.basisNsof, hfbasisdata.anDecNsof,
          hfbasisdata.usfbScfBasis,
          vWfComb
          );
      Real const rNC =
        ComputeNormalizationCoefficient(
          hfbasisdata.usfbScfBasis, vWfCombScf
          );
      ColumnVector const vWf = vWfComb / rNC;
      SetMatrixColumn( mxWf, iOrbital, vWf );
      ColumnVector const vWfScf = vWfCombScf / rNC;
      ColumnVector const vWfZSquare =
        InsertElementToColumnVector(
          ConvertVectorFromScfToStandardBasis(
            wv, hfbasisdata.basisSof, hfbasisdata.usfbScfBasis,
            product( vWfScf, vWfScf )
            ),
          hfbasisdata.iIndexToZ,
          0.0
          );
#ifdef DEBUG_OUTPUT
      cout << "  Computing operators J^k_{nl}..." << endl;
#endif
      Integer const nJCountCurr = hfatomicdata.anJCounts( iOrbital );
      if( nJCountCurr > amxOpsZSlater.length() )
        {
        OwnError( "DoNondiagNeglHFComputation: invalid number of J operators" );
        return oResults;
        }
      Array< Matrix > amxOpJCurr( nJCountCurr );
      for( Integer iJ = 0; iJ < nJCountCurr; iJ++ )
        {
        ColumnVector const vY = amxOpsZSlater( iJ ) * vWfZSquare;
        ColumnVector const vYScf =
          ConvertVectorToNearestScfBasis(
            wv, hfbasisdata.basisZNsof, hfbasisdata.anDecZNsof, 
            hfbasisdata.usfbScfBasis, vY
            );
        Matrix const mxOpZJ =
          ComputeOperator(
            wv,
            ProductNsofBlockComputator(
              wv, hfbasisdata.usfbScfBasis, vYScf
              ),
            hfbasisdata.basisZNsof,
            hfbasisdata.anDecZNsof
            );
        amxOpJCurr( iJ ) =
          RemoveZFromMatrix( mxOpZJ, hfbasisdata.iIndexToZ );
        }
      aamxOpJ( iOrbital ) = amxOpJCurr;
#ifdef DEBUG_OUTPUT
      cout << "  Computing operators K^k_{nl}..." << endl;
#endif
      Integer const nKCountCurr = hfatomicdata.anKCounts( iOrbital );
      if( nKCountCurr > amxOpsZSlater.length() )
        {
        OwnError( "DoNondiagNeglHFComputation: invalid number of K operators" );
        return oResults;
        }
      Array< Matrix > amxOpKCurr( nKCountCurr );
      for( Integer iK = 0; iK < nKCountCurr; iK++ )
        {
        Matrix const mxOpZProd =
          ComputeOperator(
            wv,
            ProductNsofBlockComputator(
              wv, hfbasisdata.usfbScfBasis, vWfScf
              ),
            hfbasisdata.basisZNsof,
            hfbasisdata.anDecZNsof
            );
        Matrix const mxOpZSlater = amxOpsZSlater( iK );
        Matrix const mxOpZK = mxOpZProd * mxOpZSlater * mxOpZProd;
        amxOpKCurr( iK ) = 
          RemoveZFromMatrix( mxOpZK, hfbasisdata.iIndexToZ );
        }
      aamxOpK( iOrbital ) = amxOpKCurr;
      logger.NotifyOrbitalInfo( 
        iOrbital, nPQN, nAMQN, wfinfo.rEpsilon, wfinfo.anRejectionReasons
        );
      }
#ifdef DEBUG_OUTPUT
    cout << "Computing operators J_{nl}..." << endl;
#endif
    for( Integer iOrbitalEq = 0;
         iOrbitalEq < nOrbitals;
         iOrbitalEq++ )
      {
      Matrix mxOpJ( nBasis, nBasis, 0.0 );
      for( Integer iOrbitalY = 0;
           iOrbitalY < nOrbitals;
           iOrbitalY++ )
        {
        ColumnVector const vCD =
          hfatomicdata.avCD( iOrbitalEq, iOrbitalY );
        Integer const nD = vCD.length();
        if( nD > aamxOpJ( iOrbitalY ).length() )
          {
          OwnError( "DoNondiagNeglHFComputation: too long vector in avCD" );
          return oResults;
          }
        for( Integer iSlater = 0; iSlater < nD; iSlater++ )
          {
          mxOpJ += vCD( iSlater ) * ( aamxOpJ( iOrbitalY ) )( iSlater );
          }
        }
      amxOpJ( iOrbitalEq ) = mxOpJ;
      }
#ifdef DEBUG_OUTPUT
    cout << "Computing operators K_{nl}..." << endl;
#endif
    for( Integer iOrbitalEq = 0;
         iOrbitalEq < nOrbitals;
         iOrbitalEq++ )
      {
      Matrix mxOpK( nBasis, nBasis, 0.0 );
      for( Integer iOrbitalY = 0;
           iOrbitalY < nOrbitals;
           iOrbitalY++ )
        {
        ColumnVector const vCE =
          hfatomicdata.avCE( iOrbitalEq, iOrbitalY );
        Integer const nE = vCE.length();
        if( nE > aamxOpK( iOrbitalY ).length() )
          {
          OwnError( "DoNondiagNeglHFComputation: too long vector in avCE" );
          return oResults;
          }
        for( Integer iSlater = 0; iSlater < nE; iSlater++ )
          {
          mxOpK += vCE( iSlater ) * ( aamxOpK( iOrbitalY ) )( iSlater );
          }
        }
      amxOpK( iOrbitalEq ) = mxOpK;
      }
    if( nIteration == 0 )
      {
      vEpsilonsHydr = vEpsilons;
      mxWfHydr = mxWf;
      }
    ColumnVector vConvIndic( nOrbitals, 0.0 );
    for( Integer iOrbital2 = 0; iOrbital2 < nOrbitals; iOrbital2++ )
      {
      vConvIndic( iOrbital2 ) =
        ComputeMatrixNorm(
          amxOpF( iOrbital2 ) - amxOpFPrev( iOrbital2 ), 
          nMatrixNormType
          );
      }
    logger.NotifyConvergenceIndicators( vConvIndic );
    if( GetColumnVectorInfinityNorm( vConvIndic  ) < rConvThreshold )
      {
      fConvergenceObtained = true;
#ifdef DEBUG_OUTPUT
      cout << "Convergence obtained in HF iteration. Iteration stopped."
           << endl;
#endif
      break;
      }
    amxOpFPrev = amxOpF;
#ifdef DEBUG_OUTPUT
    cout << "Step: " << nIteration << ": Done." << endl;
#endif
    }
  oResults.errhf =
    fConvergenceObtained ? errhfNoError : errhfNoConvergence;
  oResults.nErrorInOrbital = -1;
  oResults.vDiagLagrangeMultipliers = vEpsilons;
  oResults.vHydrDiagLagrangeMultipliers = vEpsilonsHydr;
  oResults.mxWavefunctions = mxWf;
  oResults.mxHydrWavefunctions = mxWfHydr;
  oResults.amxOpsDirectInt = amxOpJ;
  oResults.amxOpsExchangeInt = amxOpK;
  return oResults;
  }
