
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Class for Helium HF computation (member functions)

// Implementation file

// Copyright (C) 2004  Tommi Hynlnmaa

#include "HeliumHFComputation.h"
#include "ConstructMotherScalingFunction.h"
#include "ComputeHFBasisData.h"
#include "ComputeAtomEnergy.h"
#include "VectorUtils.h"
#include "ConvertVectorToNearestScfBasis.h"
#include "DoHeliumHFComputation.h"
#include <cstdlib>

Integer const HeliumHFComputation::nAtomicNumber;
Integer const HeliumHFComputation::nOrbitals;

HeliumHFComputation::HeliumHFComputation( )
  : sRunName( "UNDEFINED_RUN" ),
    anOrbitalOccupations( 1 ),
    nMaxIterations( 0 ),
    rConvThreshold( 0.0 ),
    siScfResolution( 0 ),
    nConvNormType( mxnormOne ),
    nExitCode( 1 ),
    pOutputHandler( NULL ),
    pLogger( NULL )
  {
  anOrbitalOccupations( 0 ) = 2;
  }

void HeliumHFComputation::SetRunName( string sRunName1 )
  {
  sRunName = sRunName1;
  }

void HeliumHFComputation::SetParameters(
  Integer const nMaxIterations1,
  Real rConvThreshold1,
  ScalingIndex siScfResolution1,
  EnumMatrixNormType nConvNormType1
  )
  {
  nMaxIterations = nMaxIterations1;
  rConvThreshold = rConvThreshold1;
  siScfResolution = siScfResolution1;
  nConvNormType = nConvNormType1;
  }

void HeliumHFComputation::InitBasis( )
  {
  pair< Basis, Integer > const basis_i = GetBasis( );
  hfbasisdata = ComputeHFBasisData( wv, basis_i.first, basis_i.second );
  }

void HeliumHFComputation::InitMotherScalingFunction( )
  {
  usfbfnScf =
    ConstructMotherScalingFunction( wv, siScfResolution );
  motherintegral = MotherIntegral(
    usfbfnScf.RK4Integrate().ComputeFunctionAtIntegerPerTwo(),
    1.0
    );  
  }

void HeliumHFComputation::InitOrbitalParams( )
  {
  hforbparams1s = GetOrbitalParams( );
  }

void HeliumHFComputation::Initialize( )
  {
  pOutputHandler = GetOutputHandler( );
  pLogger = GetLogger( );
  pOutputHandler->SetRunName( sRunName );
  pOutputHandler->Start( );
  pOutputHandler->StartInitialization( );
  InitFilters( );
  InitBasis( );
  InitMotherScalingFunction( );
  InitOrbitalParams( );
  pOutputHandler->SetBasisData( hfbasisdata );
  pOutputHandler->EndInitialization( );
  }

Triplet< Real, ColumnVector, ColumnVector > HeliumHFComputation::XComputeAtomEnergy(
  HeliumHFComputationResults const & oResults
  ) const
  {
  Integer const nBasis = hfbasisdata.basisSof.length( );
  Matrix const mxWavefunctions =
    Matrix( oResults.vWf1s );
  ColumnVector const vDiagLagrangeMultipliers =
    GetColumnVector( oResults.rEpsilon1s );
  Array< Matrix > amxOpsDirectInt( 1 );
  amxOpsDirectInt( 0 ) = oResults.mxOpJ1s;
  Array< Matrix > amxOpsExchangeInt( 1 );
  amxOpsExchangeInt( 0 ) = Matrix( nBasis, nBasis, 0.0 );
  Triplet< Real, ColumnVector, ColumnVector > const
    rEnergy_vJ_vK =
    ComputeAtomEnergy(
      wv,
      hfbasisdata.basisSof,
      hfbasisdata.basisNsof,
      hfbasisdata.anDecNsof,
      hfbasisdata.usfbScfBasis,
      anOrbitalOccupations,
      mxWavefunctions,
      vDiagLagrangeMultipliers,
      amxOpsDirectInt,
      amxOpsExchangeInt
      );
  return rEnergy_vJ_vK;
  }

HeliumHFComputationResults
HeliumHFComputation::DoComputation( ) const
  {
  return
    DoHeliumHFComputation(
      wv,
      hfbasisdata.basisSof,
      hfbasisdata.anDecSof,
      hfbasisdata.basisNsof,
      hfbasisdata.anDecNsof,
      hfbasisdata.basisZNsof,
      hfbasisdata.anDecZNsof,
      hfbasisdata.iIndexToZ,
      hfbasisdata.usfbScfBasis,
      motherintegral,
      vD1Filter,
      vD2Filter,
      nAtomicNumber,
      nMaxIterations,
      rConvThreshold,
      nConvNormType,
      hforbparams1s,
      *pLogger
      );
  }

void HeliumHFComputation::Run( )
  {
  pOutputHandler->StartComputation( );
  HeliumHFComputationResults const oResults = DoComputation( );
  pOutputHandler->EndComputation( );
  pOutputHandler->NotifyResults( oResults );
  if( oResults.errhf == errhfNoError )
    {
    Triplet< Real, ColumnVector, ColumnVector > const rEnergy_vJ_vK =
      XComputeAtomEnergy( oResults );
    pOutputHandler->NotifyEnergy(
      rEnergy_vJ_vK.first, rEnergy_vJ_vK.second, rEnergy_vJ_vK.third
      );
    ColumnVector const vWfScf =
      ConvertVectorToNearestScfBasis(
        wv, hfbasisdata.basisNsof, hfbasisdata.anDecNsof,
        hfbasisdata.usfbScfBasis,
        oResults.vWf1s
        );
    ColumnVector const vWfHydrScf =
      ConvertVectorToNearestScfBasis(
        wv, hfbasisdata.basisNsof, hfbasisdata.anDecNsof,
        hfbasisdata.usfbScfBasis,
        oResults.vWf1sHydr
        );
    pOutputHandler->NotifyScfWavefunction( vWfScf );
    pOutputHandler->NotifyScfHydrWavefunction( vWfHydrScf );
    }
  nExitCode = 
    ( oResults.errhf == errhfNoError ) ? 0 : 1;
  pOutputHandler->Report( );
  pOutputHandler->Save( );
  pOutputHandler->Done( );
  }

int HeliumHFComputation::GetExitCode( ) const
  {
  return nExitCode;
  }

HeliumHFComputation::~HeliumHFComputation( )
  {
  if( pOutputHandler != NULL )
    {
    delete pOutputHandler;
    pOutputHandler = NULL;
    }
  if( pLogger != NULL )
    {
    delete pLogger;
    pLogger = NULL;
    }
  }
