
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Matrix utility functions.

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "MatrixUtils.h"
#include "OwnError.h"

Matrix
RemoveRowFromMatrix(
  Matrix mx,
  Integer iRow
  )
  {
  Integer const nRows = mx.rows();
  Integer const nColumns = mx.columns();
  if( iRow < 0 || iRow >= nRows )
    {
    OwnError( "RemoveRowFromMatrix: index out of range" );
    return Matrix( 0, 0 );
    }
  if( nRows == 0 )
    {
    return mx;
    }
  if( nColumns == 0 )
    {
    return Matrix( nRows - 1, 0 );
    }
  return
    ( iRow < nRows - 1 )
    ? mx.extract( 0, 0, iRow - 1, nColumns - 1 ).stack(
      mx.extract( iRow + 1, 0, nRows - 1, nColumns - 1 )
      )
    : ( ( iRow > 0 ) 
        ? mx.extract( 0, 0, iRow - 1, nColumns - 1 )
        : Matrix( 0, nColumns ) );
  }

Matrix
RemoveColumnFromMatrix(
  Matrix mx,
  Integer iColumn
  )
  {
  Integer const nRows = mx.rows();
  Integer const nColumns = mx.columns();
  if( iColumn < 0 || iColumn >= nColumns )
    {
    OwnError( "RemoveColumnFromMatrix: index out of range" );
    return Matrix( 0, 0 );
    }
  if( nColumns == 0 )
    {
    return mx;
    }
  if( nRows == 0 )
    {
    return Matrix( 0, nColumns - 1 );
    }
  return
    ( iColumn < nColumns - 1 )
    ? mx.extract( 0, 0, nRows - 1, iColumn - 1 ).append(
      mx.extract( 0, iColumn + 1, nRows - 1, nColumns - 1 )
      )
    : ( ( iColumn > 0 ) 
        ? mx.extract( 0, 0, nRows - 1, iColumn - 1 )
        : Matrix( nRows, 0 ) );
  }

Matrix
InsertColumnToMatrix(
  Matrix mx,
  ColumnVector v,
  Integer i
  )
  {
  Integer const nRows = mx.rows();
  Integer const nColumns = mx.columns();
  if( v.length() != nRows )
    {
    OwnError( "InsertColumnToMatrix: invalid vector size" );
    return Matrix( 0, 0 );
    }
  // i may be equal to nColumns
  if( i < 0 || i > nColumns )
    {
    OwnError( "InsertColumnToMatrix: index out of range" );
    return Matrix( 0, 0 );
    }
  if( nColumns == 0 )
    {
    return Matrix( v );
    }
  if( nRows == 0 )
    {
    return Matrix( 0, nColumns - 1 );
    }
  return
    ( i > 0 )
    ? ( mx.extract( 0, 0, nRows - 1, i - 1 )
        .append( Matrix( v ) )
        .append(
          mx.extract( 0, i, nRows - 1, nColumns - 1 )
          )
        )
    : Matrix( v ).append( mx );
  }

Matrix
InsertRowToMatrix(
  Matrix mx,
  RowVector v,
  Integer i
  )
  {
  Integer const nRows = mx.rows();
  Integer const nColumns = mx.columns();
  if( v.length() != nColumns )
    {
    OwnError( "InsertRowToMatrix: invalid vector size" );
    return Matrix( 0, 0 );
    }
  // i may be equal to nRows
  if( i < 0 || i > nRows )
    {
    OwnError( "InsertRowToMatrix: index out of range" );
    return Matrix( 0, 0 );
    }
  if( nRows == 0 )
    {
    return Matrix( v );
    }
  if( nColumns == 0 )
    {
    return Matrix( nRows - 1, 0 );
    }
  return
    ( i > 0 )
    ? ( mx.extract( 0, 0, i - 1, nColumns - 1 )
        .stack( Matrix( v ) )
        .stack(
          mx.extract( i, 0, nRows - 1, nColumns - 1 )
          )
        )
    : Matrix( v ).stack( mx );
  }


Matrix
InsertZToMatrix(
  Matrix mx,
  Integer i
  )
  {
  return
    InsertColumnToMatrix(
      InsertRowToMatrix(
        mx, RowVector( mx.columns(), 0.0 ), i
        ),
      ColumnVector( mx.rows() + 1, 0.0 ),
      i
      );
  }

Matrix
RemoveZFromMatrix(
  Matrix mx,
  Integer i
  )
  {
  return
    RemoveRowFromMatrix(
      RemoveColumnFromMatrix(
        mx, i
        ),
      i
      );
  }

void
SetMatrixColumn(
  Matrix & mx,
  Integer i,
  ColumnVector v
  )
  {
  Integer const nRows = mx.rows();
  Integer const nColumns = mx.columns();
  if( i < 0 || i >= nColumns )
    {
    OwnError( "SetMatrixColumn: index out of range" );
    return;
    }
  if( v.length() != nRows )
    {
    OwnError( "SetMatrixColumn: invalid vector length" );
    return;
    }
  for( Integer i2 = 0; i2 < nRows; i2++ )
    {
    mx( i2, i ) = v( i2 );
    }
  }

void
SetMatrixRow(
  Matrix & mx,
  Integer i,
  RowVector v
  )
  {
  Integer const nRows = mx.rows();
  Integer const nColumns = mx.columns();
  if( i < 0 || i >= nRows )
    {
    OwnError( "SetMatrixRow: index out of range" );
    return;
    }
  if( v.length() != nColumns )
    {
    OwnError( "SetMatrixRow: invalid vector length" );
    return;
    }
  for( Integer i2 = 0; i2 < nColumns; i2++ )
    {
    mx( i, i2 ) = v( i2 );
    }
  }
