
// Distributed under GNU Lesser General Public License
// See file LICENSE

// Class for HF computation neglecting 
// nondiagonal Lagrange multipliers
// (member functions)

// Implementation file

// Copyright (C) 2004  Tommi Hynlnmaa

#include "NNHFComputation.h"
#include "ConstructMotherScalingFunction.h"
#include "ComputeHFBasisData.h"
#include "ComputeAtomEnergy.h"
#include "VectorUtils.h"
#include "DoNondiagNeglHFComputation.h"
#include "ComputeScfWavefunctions.h"
#include "ComputeInnerProducts.h"
#include <cstdlib>

NNHFComputation::NNHFComputation( )
  : sRunName( "UNDEFINED_RUN" ),
    nMaxIterations( 0 ),
    rConvThreshold( 0.0 ),
    siScfResolution( 0 ),
    nConvNormType( mxnormOne ),
    nExitCode( 1 ),
    pOutputHandler( NULL ),
    pLogger( NULL )
  {
  }

void NNHFComputation::SetRunName( string sRunName1 )
  {
  sRunName = sRunName1;
  }

void NNHFComputation::SetParameters(
  Integer const nMaxIterations1,
  Real rConvThreshold1,
  ScalingIndex siScfResolution1,
  EnumMatrixNormType nConvNormType1
  )
  {
  nMaxIterations = nMaxIterations1;
  rConvThreshold = rConvThreshold1;
  siScfResolution = siScfResolution1;
  nConvNormType = nConvNormType1;
  }

void NNHFComputation::InitBasis( )
  {
  pair< Basis, Integer > const basis_i = GetBasis( );
  hfbasisdata = ComputeHFBasisData( wv, basis_i.first, basis_i.second );
  }

void NNHFComputation::InitMotherScalingFunction( )
  {
  usfbfnScf =
    ConstructMotherScalingFunction( wv, siScfResolution );
  motherintegral = MotherIntegral(
    usfbfnScf.RK4Integrate( ).ComputeFunctionAtIntegerPerTwo(),
    1.0
    );  
  }

void NNHFComputation::InitAtomicData( )
  {
  hfatomicdata = GetAtomicData( );
  }

void NNHFComputation::InitOrbitalParams( )
  {
  aOrbitalParams = GetOrbitalParams( );
  }

void NNHFComputation::Initialize( )
  {
  pOutputHandler = GetOutputHandler( );
  pLogger = GetLogger( );
  pOutputHandler->SetRunName( sRunName );
  pOutputHandler->Start( );
  pOutputHandler->StartInitialization( );
  InitFilters( );
  InitBasis( );
  InitMotherScalingFunction( );
  InitAtomicData( );
  pOutputHandler->SetAtomicData( hfatomicdata );
  pOutputHandler->SetBasisData( hfbasisdata );
  InitOrbitalParams( );
  pOutputHandler->EndInitialization( );
  }

Triplet< Real, ColumnVector, ColumnVector > NNHFComputation::XComputeAtomEnergy(
  NondiagNeglHFComputationResults const & oResults
  ) const
  {
  Triplet< Real, ColumnVector, ColumnVector > const
    rEnergy_vJ_vK =
    ComputeAtomEnergy(
      wv,
      hfbasisdata.basisSof,
      hfbasisdata.basisNsof,
      hfbasisdata.anDecNsof,
      hfbasisdata.usfbScfBasis,
      hfatomicdata.anOccupations,
      oResults.mxWavefunctions,
      oResults.vDiagLagrangeMultipliers,
      oResults.amxOpsDirectInt,
      oResults.amxOpsExchangeInt
      );
  return rEnergy_vJ_vK;
  }

NondiagNeglHFComputationResults
NNHFComputation::DoComputation( ) const
  {
  return
    DoNondiagNeglHFComputation(
      hfatomicdata,
      hfbasisdata,
      wv,
      motherintegral,
      vD1Filter,
      vD2Filter,
      nMaxIterations,
      rConvThreshold,
      nConvNormType,
      aOrbitalParams,
      *pLogger
      );
  }

void NNHFComputation::Run( )
  {
  pOutputHandler->StartComputation( );
  NondiagNeglHFComputationResults const oResults = DoComputation( );
  pOutputHandler->EndComputation( );
  pOutputHandler->NotifyResults( oResults );
  if( oResults.errhf == errhfNoError )
    {
    Triplet< Real, ColumnVector, ColumnVector > const rEnergy_vJ_vK =
      XComputeAtomEnergy( oResults );
    pOutputHandler->NotifyEnergy(
      rEnergy_vJ_vK.first, rEnergy_vJ_vK.second, rEnergy_vJ_vK.third
      );
    Matrix const mxWfScf = 
      ComputeScfWavefunctions( wv, hfbasisdata, oResults.mxWavefunctions );
    Matrix const mxWfHydrScf = 
      ComputeScfWavefunctions( wv, hfbasisdata, oResults.mxHydrWavefunctions );
    Matrix const mxInnerProducts =
      ComputeInnerProducts( hfbasisdata.usfbScfBasis, mxWfScf );
    pOutputHandler->NotifyInnerProducts( mxInnerProducts );
    pOutputHandler->NotifyScfWavefunctions( mxWfScf );
    pOutputHandler->NotifyScfHydrWavefunctions( mxWfHydrScf );
    }
  nExitCode = 
    ( oResults.errhf == errhfNoError ) ? 0 : 1;
  pOutputHandler->Report( );
  pOutputHandler->Save( );
  pOutputHandler->Done( );
  }

int NNHFComputation::GetExitCode( ) const
  {
  return nExitCode;
  }

NNHFComputation::~NNHFComputation( )
  {
  if( pOutputHandler != NULL )
    {
    delete pOutputHandler;
    pOutputHandler = NULL;
    }
  if( pLogger != NULL )
    {
    delete pLogger;
    pLogger = NULL;
    }
  }
