
// Distributed under GNU Lesser General Public License
// See file LICENSE


// class ScalingFunction

// Scaling function of interpolating wavelets.

// A special kind of tabulated function, where some 
// mathematical properties of the scaling function are 
// used for computations.

// Implements computation of multi-indexed basis function
// values.

// Header file without implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#ifndef ScalingFunction_h
#define ScalingFunction_h

#include "OctaveIncludes.h"
#include "CommonTypes.h"
#include "TabulatedFunction.h"
#include "BasisMultiIndex.h"
#include "NTwoToIntegerPower.h"
#include <cmath>

class ScalingFunction : public TabulatedFunction
  {
    protected:

  Integer k;
  Integer n;

    public:

  ScalingFunction( ColumnVector x1,
		   ColumnVector y1,
		   Integer points1,
		   Real step1,
		   Integer k1,
		   Integer n1 )
    : TabulatedFunction( x1, y1, points1, step1),
    k( k1 ), n( n1 )
      {
      }

  Real MotherScfValue( Real xarg ) const
      {
      Real result = 0.0;
      if( xarg >= x( 0 ) && xarg <= x( x.length() - 1 ) )
        {
        Integer const i = 
          n + Integer( floor( NTwoToIntegerPower( k ) * xarg ) );
        if( i >= 0 && i < points - 1 ) 
          {
          Real const x1 = x( i );
          Real const x2 = x( i + 1 );
          Real const y1 = y( i );
          Real const y2 = y( i + 1 );
          result = ( ( x2 - xarg ) * y1 + ( xarg - x1 ) * y2 ) 
            / ( x2 - x1 );
          }
        }
      return result;
      }

  Real ScfValue( ScalingIndex k, TranslationIndex i, Real xarg ) const
      {
      return MotherScfValue( NTwoToIntegerPower( k ) * xarg - i );
      }

  Real WaveletValue( ScalingIndex k, TranslationIndex i, Real xarg ) const
      {
      Real x2;
      x2 = NTwoToIntegerPower( k ) * xarg - i;
      return ( - MotherScfValue( 2 * x2 - 1 ) );
      }

  Real BasisFunctionValue( BasisMultiIndex alpha, Real xarg ) const
      {
      return 
        ( alpha.wt == 0 )
        ? ScfValue( alpha.si, alpha.ti, xarg )
        : WaveletValue( alpha.si, alpha.ti, xarg );
      }

  };

#endif
