
// Distributed under GNU Lesser General Public License
// See file LICENSE


// class SlaterIntegralFunction

// Class for representing a Slater integral
// as a tabulated function.

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "SlaterIntegralFunction.h"
#include "ComputeSlaterIntegral.h"
#include "GetLinearSpace.h"
#include "OwnError.h"
#include <cmath>
#include <cstdio>

SlaterIntegralFunction::SlaterIntegralFunction(
  ScalingFunction const & scf,
  Basis const & basis,
  Integer nSlaterK,
  ColumnVector vComponents1,
  ColumnVector vComponents2,
  Real rRange,
  Integer nSlaterIntPoints,
  bool fNegativePartSignChange
  )
  {
  if( ( nSlaterIntPoints < 2 ) ||
      ( nSlaterIntPoints % 2 == 0 ) )
    {
    OwnError( "SlaterIntegralFunction::SlaterIntegralFunction(..): invalid number of points" );
    }
  x = GetLinearSpace( -rRange, rRange, nSlaterIntPoints );
  step = x( 1 ) - x( 0 );
  points = nSlaterIntPoints;
  y = ComputeSlaterIntegral(
    scf, basis, nSlaterK,
    vComponents1, vComponents2,
    rRange, nSlaterIntPoints,
    fNegativePartSignChange,
    y_a,
    y_b
    );
  }

Real SlaterIntegralFunction::ValueInRangeAt( Real xarg ) const
  {
  Real result = 0.0;
  if( points == 0 )
    {
    result = 0.0;
    }
  else if( xarg < x( 0 ) || xarg > x( x.length() - 1 ) )
    {
    OwnError( "SlaterIntegralFunction::ValueInRangeAt: r out of range (1)" );
    result = 0.0;
    }
  else
    {
    Integer const i = Integer( floor( ( xarg - x( 0 ) ) / step ) );
    if( i < 0 || i > points - 1 ) 
      {
      OwnError( "SlaterIntegralFunction::ValueInRangeAt: r out of range (2)" );
      }
    else if( i == points - 1 )
      {
      result = y( i );
      }
    else
      {
      Real const x1 = x( i );
      Real const x2 = x( i + 1 );
      Real const y1 = y( i );
      Real const y2 = y( i + 1 );
      result = ( ( x2 - xarg ) * y1 + ( xarg - x1 ) * y2 ) / ( x2 - x1 );
      }
    }
  return result;
  }

Real SlaterIntegralFunction::SlaterIntegralValueAt( Real rarg ) const
  {
  Real result = 1E30;
  Real r = fabs( rarg );
  if( points == 0 )
    {
    result = 0.0;
    }
  else if( r < 0.0 )
    {
    result = 0.0;
    }
  else if( r < x( 0 ) )
    {
    result = ValueBelow( r );
    }
  else if( r >= x( x.length() - 1 ) )  
    {
    result = ValueAbove( r );
    }
  else
    {
    result = ValueInRangeAt( r );
    }
  return result;
  }

bool SlaterIntegralFunction::CheckContinuity() const
  {
  Real rmin = x( 0 );
  Real rmax = x( x.length() - 1 );
  Real y_rmin_1 = ValueBelow( rmin );
  Real y_rmin_2 = ValueInRangeAt( rmin );
  Real y_rmax_1 = ValueAbove( rmax );
  Real y_rmax_2 = ValueInRangeAt( rmax );
  bool cont_at_rmin = 
    ( fabs( y_rmin_1 - y_rmin_2 ) < 1E-10 );
  bool cont_at_rmax = 
    ( fabs( y_rmax_1 - y_rmax_2 ) < 1E-10 );
  bool result = ( cont_at_rmin && cont_at_rmax );
  if( ! result )
    {
    fprintf( stderr, "Y(r_min_1) = %G\n", y_rmin_1 );
    fprintf( stderr, "Y(r_min_2) = %G\n", y_rmin_2 );
    fprintf( stderr, "Y(r_max_1) = %G\n", y_rmax_1 );
    fprintf( stderr, "Y(r_max_2) = %G\n", y_rmax_2 );
    OwnError( "SlaterIntegralFunction: not continuous" );
    }
  return result;
  }

bool SlaterIntegralFunction::CheckConsistency() const
  {
  bool result = false;
  if( points == 0 )
    {
    result = true;
    }
  else 
    {
    if( ( points >= 2 ) &&
        ( x.length() == points ) &&
        ( y.length() == points ) &&
        ( step > 0.0 ) &&
        ( x( 0 ) >= 0.0 ) &&
        ( fabs( ( x( x.length() - 1 ) - x( 0 ) ) / ( points - 1 ) 
                - step ) < 1E-10 ) &&
        ( fabs( ( x( 1 ) - x( 0 ) ) - step ) < 1E-10 ) )
      {
      if( CheckContinuity() )
        {
        result = true;
        }
      else
        {
        result = false;
        }
      }
    else
      {
      OwnError( "SlaterIntegralFunction: error in tabulation" );
      result = false;
      }
    }
  return result;
  }
