
// Distributed under GNU Lesser General Public License
// See file LICENSE


// class SlaterIntegralFunction

// Class for representing a Slater integral
// as a tabulated function.

// Computation of values of the function is
// implemeted also outside tabulation interval
// (in whole nonnegative real axis).

// Header file

// Copyright (C) 2003  Tommi Hynlnmaa

#ifndef SlaterIntegralFunction_h
#define SlaterIntegralFunction_h

#include "OctaveIncludes.h"
#include "TabulatedFunction.h"
#include "RealToNonnegIntPower.h"
#include "ScalingFunction.h"
#include "Basis.h"

class SlaterIntegralFunction : public TabulatedFunction
  {
    protected:
  
  Integer slater_k;
  Real y_a;
  Real y_b;

    public:

  SlaterIntegralFunction()
    : slater_k( 0 ), y_a( 0.0 ), y_b( 0.0 )
      {
      }

  SlaterIntegralFunction( ColumnVector x1,
                          ColumnVector y1,
                          Integer points1,
                          Real step1,
                          Integer slater_k1,
                          Real y_a1,
                          Real y_b1 )
    : TabulatedFunction( x1, y1, points1, step1 ),
    slater_k( slater_k1 ),
    y_a( y_a1 ), y_b( y_b1 )
      {
      }

  SlaterIntegralFunction(
    ScalingFunction const & scf,
    Basis const & basis,
    Integer nSlaterK,
    ColumnVector vComponents1,
    ColumnVector vComponents2,
    Real rRange,
    Integer nSlaterIntPoints,
    bool fNegativePartSignChange
    );

  Real ValueBelow( Real xarg ) const
      {
      return RealToNonnegIntPower( xarg, slater_k ) * y_a;
      }

  Real ValueAbove( Real xarg ) const
      {
      return ( 1.0 / RealToNonnegIntPower( xarg, slater_k + 1 ) ) * y_b;
      }

  Real ValueInRangeAt( Real xarg ) const;

  Real SlaterIntegralValueAt( Real rarg ) const;

  bool CheckContinuity() const;

  bool CheckConsistency() const;

  bool IsEmpty() const
      {
      return ( points == 0 );
      }

  Integer GetSlaterK() const
      {
      return slater_k;
      }

  Real GetYa() const
      {
      return y_a;
      }

  Real GetYb() const
      {
      return y_b;
      }

  };

#ifdef OCTAVE_DL_COMPILATION
#include "SlaterIntegralFunction.cc"
#endif

#endif
