
// Distributed under GNU Lesser General Public License
// See file LICENSE


// Class UsfbTabulatedFunction

// Class for a function whose values are tabulated
// at x values of a uniform basis.

// Implementation file

// Copyright (C) 2003  Tommi Hynlnmaa

#include "UsfbTabulatedFunction.h"
#include "ArrayUtils.h"
#include "RK4IntegrateValues.h"
#include "NTwoToIntegerPower.h"
#include "OwnError.h"
#include <cmath>

UsfbTabulatedFunction
UsfbTabulatedFunction::RK4Integrate() const
  {
  if( usfb.si <= 0 )
    {
    OwnError( "UsfbTabulatedFunction::RK4Integrate: invalid resolution" );
    return UsfbTabulatedFunction();
    }
  Integer const nOffsetMin =
    ( usfb.tiMin % 2 != 0 ) ? 1 : 0;
  Integer const nOffsetMax =
    ( usfb.tiMax % 2 != 0 ) ? 1 : 0;
  Integer const nPoints = vValues.length();
  ColumnVector const vValuesForInt =
    ColumnVector(
      GetRangeFromArray(
        vValues, nOffsetMin, nPoints - 1 - nOffsetMax
        ) 
      );
  Real const rStep = 1.0 / NTwoToIntegerPower( usfb.si - 1 );
  UsfbTabulatedFunction const usfbfnInt =
    UsfbTabulatedFunction(
      UniformBasis(
        0, 
        usfb.si - 1,
        ( usfb.tiMin + nOffsetMin ) / 2,
        ( usfb.tiMax - nOffsetMax ) / 2
        ),
      RK4IntegrateValues( vValuesForInt, rStep )
      );
  if( ! usfbfnInt.IsCorrect() )
    {
    OwnError( "UsfbTabulatedFunction::RK4Integrate: internal error" );
    return UsfbTabulatedFunction();
    }
  return usfbfnInt;
  }

UsfbTabulatedFunction 
UsfbTabulatedFunction::ComputeFunctionAtIntegerPerTwo() const
  {
  if( usfb.si <= 0 )
    {
    OwnError( "UsfbTabulatedFunction::ComputeValuesAtIntegerPerTwo: too coarse basis" );
    return UsfbTabulatedFunction();
    }
  Integer const nValues = vValues.length();
  Integer const nStep = NTwoToIntegerPower( usfb.si - 1 );
  Integer const tiMin =
    Integer( ceil( Real( usfb.tiMin ) / nStep ) );
  Integer const tiMax =
    Integer( floor( Real( usfb.tiMax ) / nStep ) );
  Integer const n = tiMax - tiMin + 1;
  if( n < 0 )
    {
    OwnError( "UsfbTabulatedFunction::ComputeFunctionAtIntegerPerTwo: internal error" );
    return UsfbTabulatedFunction();
    }
  ColumnVector vNewValues( n, 0.0 );
  for( TranslationIndex ti = tiMin; ti <= tiMax; ti++ )
    {
    Integer const iValue = nStep * ti - usfb.tiMin;
    if( iValue < 0 || iValue >= nValues )
      {
      OwnError( "UsfbTabulatedFunction::ComputeFunctionAtIntegerPerTwo: index out of range" );
      return UsfbTabulatedFunction();
      }
    vNewValues( ti - tiMin ) = vValues( iValue );
    }
  UsfbTabulatedFunction const usfbfnResult =
    UsfbTabulatedFunction(
      UniformBasis( 0, 1, tiMin, tiMax ),
      vNewValues
      );
  if( ! usfbfnResult.IsCorrect() )
    {
    OwnError( "UsfbTabulatedFunction::ComputeFunctionAtIntegerPerTwo: internal error" );
    return UsfbTabulatedFunction();
    }
  return usfbfnResult;
  }
